import binascii
from pathlib import Path
from typing import List, Union, Dict, Optional, Tuple, TYPE_CHECKING

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import Certificate
from cryptography.x509.ocsp import OCSPResponseStatus

from sslyze.plugins.certificate_info._cert_chain_analyzer import CertificateDeploymentAnalysisResult
from sslyze.plugins.certificate_info._certificate_utils import get_common_names, extract_dns_subject_alternative_names

from sslyze.plugins.plugin_base import ScanCommandCliConnector, OptParseCliOption

if TYPE_CHECKING:
    from sslyze.plugins.certificate_info.implementation import CertificateInfoScanResult
    from sslyze.plugins.certificate_info.implementation import CertificateInfoExtraArgument  # noqa: F401


class _CertificateInfoCliConnector(
    ScanCommandCliConnector["CertificateInfoScanResult", "CertificateInfoExtraArgument"]
):

    _cli_option = "certinfo"
    _cli_description = "Retrieve and analyze a server's certificate(s) to verify its validity."

    @classmethod
    def get_cli_options(cls) -> List[OptParseCliOption]:
        scan_command_option = super().get_cli_options()
        scan_command_option.append(
            OptParseCliOption(
                option="certinfo_ca_file",
                help="To be used with --certinfo. Path to a file containing root certificates in PEM format that will"
                " be used to verify the validity of the server's certificate.",
                action="store",
            )
        )
        return scan_command_option

    @classmethod
    def find_cli_options_in_command_line(
        cls, parsed_command_line: Dict[str, Union[None, bool, str]]
    ) -> Tuple[bool, Optional["CertificateInfoExtraArgument"]]:
        # Avoid circular imports
        from sslyze.plugins.certificate_info.implementation import CertificateInfoExtraArgument  # noqa: F811

        # Check if --certinfo was used
        is_scan_cmd_enabled, _ = super().find_cli_options_in_command_line(parsed_command_line)

        # Check if --certinfo_ca_file was used
        extra_arguments = None
        try:
            certinfo_ca_file = parsed_command_line["certinfo_ca_file"]
            if certinfo_ca_file:
                if not isinstance(certinfo_ca_file, str):
                    raise TypeError(f"Expected a str for certinfo_ca_file but received {certinfo_ca_file}")
                extra_arguments = CertificateInfoExtraArgument(custom_ca_file=Path(certinfo_ca_file))
        except KeyError:
            pass

        return is_scan_cmd_enabled, extra_arguments

    TRUST_FORMAT = "{store_name} CA Store ({store_version}):"
    NO_VERIFIED_CHAIN_ERROR_TXT = "ERROR - Could not build verified chain (certificate untrusted?)"

    @classmethod
    def result_to_console_output(cls, result: "CertificateInfoScanResult") -> List[str]:
        result_as_txt = [cls._format_title("Certificates Information")]

        # SNI
        server_name_indication = result.hostname_used_for_server_name_indication
        result_as_txt.append(cls._format_field("Hostname sent for SNI:", server_name_indication))

        # Display each certificate deployment
        result_as_txt.append(
            cls._format_field("Number of certificates detected:", str(len(result.certificate_deployments)))
        )
        for index, cert_deployment in enumerate(result.certificate_deployments):
            result_as_txt.append("\n")
            result_as_txt.extend(cls._cert_deployment_to_console_output(index, cert_deployment))

        return result_as_txt

    @classmethod
    def _cert_deployment_to_console_output(
        cls, index: int, cert_deployment: CertificateDeploymentAnalysisResult
    ) -> List[str]:
        leaf_certificate = cert_deployment.received_certificate_chain[0]
        deployment_as_txt = [
            cls._format_subtitle(f"Certificate #{index} ( {leaf_certificate.public_key().__class__.__name__} )")
        ]

        deployment_as_txt.extend(cls._get_basic_certificate_text(leaf_certificate))

        # Trust section
        deployment_as_txt.append("")
        deployment_as_txt.append(cls._format_subtitle(f"Certificate #{index} - Trust"))

        hostname_validation_text = (
            "OK - Certificate matches server hostname"
            if cert_deployment.leaf_certificate_subject_matches_hostname
            else "FAILED - Certificate does NOT match server hostname"
        )
        deployment_as_txt.append(cls._format_field("Hostname Validation:", hostname_validation_text))

        # Path validation that was successfully tested
        for path_result in cert_deployment.path_validation_results:
            if path_result.was_validation_successful:
                # EV certs - Only Mozilla supported for now
                ev_txt = ""
                if cert_deployment.leaf_certificate_is_ev and path_result.trust_store.ev_oids:
                    ev_txt = ", Extended Validation"
                path_txt = f"OK - Certificate is trusted{ev_txt}"

            else:
                path_txt = f"FAILED - Certificate is NOT Trusted: {path_result.openssl_error_string}"

            deployment_as_txt.append(
                cls._format_field(
                    cls.TRUST_FORMAT.format(
                        store_name=path_result.trust_store.name, store_version=path_result.trust_store.version
                    ),
                    path_txt,
                )
            )

        if cert_deployment.verified_chain_has_legacy_symantec_anchor is None:
            symantec_str = cls.NO_VERIFIED_CHAIN_ERROR_TXT
        elif cert_deployment.verified_chain_has_legacy_symantec_anchor is True:
            symantec_str = "WARNING: Certificate distrusted by Google and Mozilla since 2018"
        elif cert_deployment.verified_chain_has_legacy_symantec_anchor is False:
            symantec_str = "OK - Not a Symantec-issued certificate"
        else:
            raise RuntimeError("Should never happen")
        deployment_as_txt.append(cls._format_field("Symantec 2018 Deprecation:", symantec_str))

        # Print the Common Names within the received certificate chain
        cns_in_received_chain: List[str] = [
            _get_subject_as_short_text(cert) for cert in cert_deployment.received_certificate_chain
        ]
        deployment_as_txt.append(cls._format_field("Received Chain:", " --> ".join(cns_in_received_chain)))

        # Print the Common Names within the verified certificate chain if validation was successful
        if cert_deployment.verified_certificate_chain:
            cns_in_certificate_chain = [
                _get_subject_as_short_text(cert) for cert in cert_deployment.verified_certificate_chain
            ]
            verified_chain_txt = " --> ".join(cns_in_certificate_chain)
        else:
            verified_chain_txt = cls.NO_VERIFIED_CHAIN_ERROR_TXT
        deployment_as_txt.append(cls._format_field("Verified Chain:", verified_chain_txt))

        if cert_deployment.verified_certificate_chain:
            chain_with_anchor_txt = (
                "OK - Anchor certificate not sent"
                if not cert_deployment.received_chain_contains_anchor_certificate
                else "WARNING - Received certificate chain contains the anchor certificate"
            )
        else:
            chain_with_anchor_txt = cls.NO_VERIFIED_CHAIN_ERROR_TXT
        deployment_as_txt.append(cls._format_field("Received Chain Contains Anchor:", chain_with_anchor_txt))

        chain_order_txt = (
            "OK - Order is valid"
            if cert_deployment.received_chain_has_valid_order
            else "FAILED - Certificate chain out of order!"
        )
        deployment_as_txt.append(cls._format_field("Received Chain Order:", chain_order_txt))

        if cert_deployment.verified_certificate_chain:
            sha1_text = (
                "OK - No SHA1-signed certificate in the verified certificate chain"
                if not cert_deployment.verified_chain_has_sha1_signature
                else "INSECURE - SHA1-signed certificate in the verified certificate chain"
            )
        else:
            sha1_text = cls.NO_VERIFIED_CHAIN_ERROR_TXT
        deployment_as_txt.append(cls._format_field("Verified Chain contains SHA1:", sha1_text))

        # Extensions section
        deployment_as_txt.extend(["", cls._format_subtitle(f"Certificate #{index} - Extensions")])

        # OCSP must-staple
        must_staple_txt = (
            "OK - Extension present"
            if cert_deployment.leaf_certificate_has_must_staple_extension
            else "NOT SUPPORTED - Extension not found"
        )
        deployment_as_txt.append(cls._format_field("OCSP Must-Staple:", must_staple_txt))

        # Look for SCT extension
        scts_count = cert_deployment.leaf_certificate_signed_certificate_timestamps_count
        if scts_count is None:
            sct_txt = "OK - Extension present"
        elif scts_count == 0:
            sct_txt = "NOT SUPPORTED - Extension not found"
        elif scts_count < 3:
            sct_txt = "WARNING - Only {} SCTs included but Google recommends 3 or more".format(str(scts_count))
        else:
            sct_txt = "OK - {} SCTs included".format(str(scts_count))
        deployment_as_txt.append(cls._format_field("Certificate Transparency:", sct_txt))

        # OCSP stapling
        deployment_as_txt.extend(["", cls._format_subtitle(f"Certificate #{index} - OCSP Stapling")])

        if cert_deployment.ocsp_response is None:
            deployment_as_txt.append(cls._format_field("", "NOT SUPPORTED - Server did not send back an OCSP response"))

        else:
            if cert_deployment.ocsp_response.response_status != OCSPResponseStatus.SUCCESSFUL:
                ocsp_resp_txt = [
                    cls._format_field(
                        "",
                        "ERROR - OCSP response status is not successful: {}".format(
                            cert_deployment.ocsp_response.response_status.name
                        ),
                    )
                ]
            else:
                ocsp_trust_txt = (
                    "OK - Response is trusted"
                    if cert_deployment.ocsp_response_is_trusted
                    else "FAILED - Response is NOT trusted"
                )

                ocsp_resp_txt = [
                    cls._format_field("OCSP Response Status:", cert_deployment.ocsp_response.response_status.name),
                    cls._format_field("Validation w/ Mozilla Store:", ocsp_trust_txt),
                ]

                if cert_deployment.ocsp_response.responder_key_hash:
                    ocsp_resp_txt.append(
                        cls._format_field("Responder Key Hash:", str(cert_deployment.ocsp_response.responder_key_hash))
                    )

                if cert_deployment.ocsp_response.responder_name:
                    ocsp_resp_txt.append(
                        cls._format_field(
                            "Responder Name:", cert_deployment.ocsp_response.responder_name.rfc4514_string()
                        )
                    )

                if cert_deployment.ocsp_response.response_status == OCSPResponseStatus.SUCCESSFUL:
                    ocsp_resp_txt.extend(
                        [
                            cls._format_field("Cert Status:", cert_deployment.ocsp_response.certificate_status.name),
                            cls._format_field("Cert Serial Number:", str(cert_deployment.ocsp_response.serial_number)),
                            cls._format_field(
                                "This Update:", cert_deployment.ocsp_response.this_update.date().isoformat()
                            ),
                        ]
                    )

                    # The Next Update field is optional: https://github.com/nabla-c0d3/sslyze/issues/481
                    if cert_deployment.ocsp_response.next_update is None:
                        next_update_str = "None"
                    else:
                        next_update_str = cert_deployment.ocsp_response.next_update.date().isoformat()
                    ocsp_resp_txt.append(cls._format_field("Next Update:", next_update_str))

            deployment_as_txt.extend(ocsp_resp_txt)

        # All done
        return deployment_as_txt

    @classmethod
    def _get_basic_certificate_text(cls, certificate: Certificate) -> List[str]:
        text_output = [
            cls._format_field(
                "SHA1 Fingerprint:", binascii.hexlify(certificate.fingerprint(hashes.SHA1())).decode("ascii")
            ),
            cls._format_field("Common Name:", _get_subject_as_short_text(certificate)),
            cls._format_field("Issuer:", _get_issuer_as_short_text(certificate)),
            cls._format_field("Serial Number:", str(certificate.serial_number)),
            cls._format_field("Not Before:", certificate.not_valid_before.date().isoformat()),
            cls._format_field("Not After:", certificate.not_valid_after.date().isoformat()),
            cls._format_field("Public Key Algorithm:", certificate.public_key().__class__.__name__),
        ]

        if certificate.signature_hash_algorithm:
            # The signature_hash_algorithm can be None if signature did not use separate hash (ED25519, ED448)
            # https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate.signature_hash_algorithm
            text_output.append(cls._format_field("Signature Algorithm:", certificate.signature_hash_algorithm.name))

        public_key = certificate.public_key()
        if isinstance(public_key, EllipticCurvePublicKey):
            text_output.append(cls._format_field("Key Size:", str(public_key.curve.key_size)))
            text_output.append(cls._format_field("Curve:", str(public_key.curve.name)))
        elif isinstance(public_key, RSAPublicKey):
            text_output.append(cls._format_field("Key Size:", str(public_key.key_size)))
            text_output.append(cls._format_field("Exponent:", str(public_key.public_numbers().e)))  # type: ignore
        else:
            # DSA Key? https://github.com/nabla-c0d3/sslyze/issues/314
            pass

        try:
            # Print the SAN extension if there's one
            text_output.append(
                cls._format_field(
                    "DNS Subject Alternative Names:", str(extract_dns_subject_alternative_names(certificate))
                )
            )
        except KeyError:
            pass

        return text_output


def _get_subject_as_short_text(certificate: Certificate) -> str:
    try:
        final_subject_field = _get_name_as_short_text(certificate.subject)
    except ValueError:
        # Cryptography could not parse the certificate https://github.com/nabla-c0d3/sslyze/issues/495
        final_subject_field = "Invalid Cert: Subject could not be parsed"
    return final_subject_field


def _get_issuer_as_short_text(certificate: Certificate) -> str:
    try:
        final_issuer_field = _get_name_as_short_text(certificate.issuer)
    except ValueError:
        # Cryptography could not parse the certificate https://github.com/nabla-c0d3/sslyze/issues/495
        final_issuer_field = "Invalid Cert: Issuer could not be parsed"
    return final_issuer_field


def _get_name_as_short_text(name_field: x509.Name) -> str:
    """Convert a name field returned by the cryptography module to a string suitable for displaying it to the user.
    """
    # Name_field is supposed to be a Subject or an Issuer; print the CN if there is one
    common_names = get_common_names(name_field)
    if common_names:
        # We don't support certs with multiple CNs
        return common_names[0]
    else:
        # Otherwise show the whole field
        return name_field.rfc4514_string()



from typing import Tuple

import nassl

from sslyze.plugins.session_resumption._resumption_with_id import retrieve_tls_session, _ScanJobResultEnum
from sslyze.server_connectivity import ServerConnectivityInfo


def resume_with_tls_ticket(server_info: ServerConnectivityInfo,) -> Tuple[_ScanJobResultEnum, bool]:
    """Perform one session resumption using TLS Session Tickets.
    """
    # Connect to the server and keep the TLS session
    session1 = retrieve_tls_session(server_info, should_enable_tls_ticket=True)
    try:
        # Recover the TLS ticket
        session1_tls_ticket = _extract_tls_session_ticket(session1)
    except IndexError:
        # Ticket was not assigned
        return _ScanJobResultEnum.TLS_TICKET_RESUMPTION, False

    # Try to resume that session using the TLS ticket
    session2 = retrieve_tls_session(server_info, session_to_resume=session1, should_enable_tls_ticket=True)
    try:
        # Recover the TLS ticket
        session2_tls_ticket = _extract_tls_session_ticket(session2)
    except IndexError:
        # Ticket was not assigned
        return _ScanJobResultEnum.TLS_TICKET_RESUMPTION, False

    # Finally, compare the two TLS Tickets
    if session1_tls_ticket != session2_tls_ticket:
        # The Ticket we supplied got ignored
        return _ScanJobResultEnum.TLS_TICKET_RESUMPTION, False

    return _ScanJobResultEnum.TLS_TICKET_RESUMPTION, True


def _extract_tls_session_ticket(ssl_session: nassl._nassl.SSL_SESSION) -> str:
    """Extract the TLS session ticket from a SSL session object or raises IndexError if the ticket was not set.
    """
    session_string = ((ssl_session.as_text()).split("TLS session ticket:"))[1]
    session_tls_ticket = (session_string.split("Compression:"))[0]
    return session_tls_ticket



from base64 import b64encode
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional

import pydantic
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509 import NameAttribute
from cryptography.x509.ocsp import OCSPResponseStatus
from cryptography.x509.oid import ObjectIdentifier  # type: ignore
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey

from sslyze import (
    CertificateInfoExtraArgument,
    CertificateInfoScanResult,
    CertificateDeploymentAnalysisResult,
    PathValidationResult,
    TrustStore,
)
from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.plugins.certificate_info._certificate_utils import (
    get_public_key_sha256,
    extract_dns_subject_alternative_names,
)


class _BaseModelWithOrmMode(pydantic.BaseModel):
    class Config:
        orm_mode = True


class CertificateInfoExtraArgumentAsJson(_BaseModelWithOrmMode):
    custom_ca_file: Path


CertificateInfoExtraArgumentAsJson.__doc__ = CertificateInfoExtraArgument.__doc__  # type: ignore


class _PublicKeyAsJson(_BaseModelWithOrmMode):
    algorithm: str
    key_size: Optional[int]  # None for Ed25519PublicKey and Ed448PublicKey

    # Only set if the algorithm is RSA
    rsa_e: Optional[int]
    rsa_n: Optional[int]

    # Only set if the algorithm is Elliptic Curve
    ec_curve_name: Optional[str]
    ec_x: Optional[int]
    ec_y: Optional[int]

    @classmethod
    def from_orm(cls, public_key: Any) -> "_PublicKeyAsJson":
        try:
            public_key_size = public_key.key_size  # type: ignore
        except AttributeError:
            public_key_size = None

        return cls(
            algorithm=public_key.__class__.__name__,
            key_size=public_key_size,
            # EC-only fields
            ec_curve_name=public_key.curve.name if isinstance(public_key, EllipticCurvePublicKey) else None,
            ec_x=public_key.public_numbers().x if isinstance(public_key, EllipticCurvePublicKey) else None,
            ec_y=public_key.public_numbers().y if isinstance(public_key, EllipticCurvePublicKey) else None,
            # RSA-only fields
            rsa_e=public_key.public_numbers().e if isinstance(public_key, RSAPublicKey) else None,
            rsa_n=public_key.public_numbers().n if isinstance(public_key, RSAPublicKey) else None,
        )


class _ObjectIdentifierAsJson(_BaseModelWithOrmMode):
    name: str
    dotted_string: str

    @classmethod
    def from_orm(cls, oid: ObjectIdentifier) -> "_ObjectIdentifierAsJson":
        return cls(name=oid._name, dotted_string=oid.dotted_string)


class _NameAttributeAsJson(_BaseModelWithOrmMode):
    oid: _ObjectIdentifierAsJson
    value: str
    rfc4514_string: str

    @classmethod
    def from_orm(cls, name_attribute: NameAttribute) -> "_NameAttributeAsJson":
        return cls(
            oid=_ObjectIdentifierAsJson.from_orm(name_attribute.oid),
            value=name_attribute.value,
            rfc4514_string=name_attribute.rfc4514_string(),
        )


class _X509NameAsJson(_BaseModelWithOrmMode):
    rfc4514_string: str
    attributes: List[_NameAttributeAsJson]

    @classmethod
    def from_orm(cls, name: x509.name.Name) -> "_X509NameAsJson":
        return cls(
            rfc4514_string=name.rfc4514_string(), attributes=[_NameAttributeAsJson.from_orm(attr) for attr in name]
        )


class _SubjAltNameAsJson(pydantic.BaseModel):
    dns: List[str]


class _HashAlgorithmAsJson(_BaseModelWithOrmMode):
    name: str
    digest_size: int

    @classmethod
    def from_orm(cls, hash_algorithm: hashes.HashAlgorithm) -> "_HashAlgorithmAsJson":
        return cls(name=hash_algorithm.name, digest_size=hash_algorithm.digest_size)


class _CertificateAsJson(_BaseModelWithOrmMode):
    as_pem: str
    hpkp_pin: str  # RFC 7469
    fingerprint_sha1: str
    fingerprint_sha256: str

    serial_number: int
    not_valid_before: datetime
    not_valid_after: datetime
    subject_alternative_name: _SubjAltNameAsJson

    # The signature_hash_algorithm can be None if signature did not use separate hash (ED25519, ED448)
    # https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate.signature_hash_algorithm
    signature_hash_algorithm: Optional[_HashAlgorithmAsJson]
    signature_algorithm_oid: _ObjectIdentifierAsJson

    # We may get garbage/invalid certificates that do not have a subject or an issuer, hence they can be None
    # https://github.com/nabla-c0d3/sslyze/issues/403
    subject: Optional[_X509NameAsJson]
    issuer: Optional[_X509NameAsJson]

    public_key: _PublicKeyAsJson

    @classmethod
    def from_orm(cls, certificate: x509.Certificate) -> "_CertificateAsJson":
        signature_hash_algorithm: Optional[_HashAlgorithmAsJson]
        if certificate.signature_hash_algorithm:
            signature_hash_algorithm = _HashAlgorithmAsJson.from_orm(certificate.signature_hash_algorithm)
        else:
            signature_hash_algorithm = None

        # We may get garbage/invalid certificates so we need to handle ValueErrors.
        # See https://github.com/nabla-c0d3/sslyze/issues/403 for more information
        subject_field: Optional[_X509NameAsJson]
        try:
            subject_field = _X509NameAsJson.from_orm(certificate.subject)
        except ValueError:
            subject_field = None

        issuer_field: Optional[_X509NameAsJson]
        try:
            issuer_field = _X509NameAsJson.from_orm(certificate.issuer)
        except ValueError:
            issuer_field = None

        return cls(
            as_pem=certificate.public_bytes(Encoding.PEM).decode("ascii"),
            hpkp_pin=b64encode(get_public_key_sha256(certificate)).decode("ascii"),
            fingerprint_sha1=b64encode(certificate.fingerprint(hashes.SHA1())).decode("ascii"),
            fingerprint_sha256=b64encode(certificate.fingerprint(hashes.SHA256())).decode("ascii"),
            serial_number=certificate.serial_number,
            not_valid_before=certificate.not_valid_before,
            not_valid_after=certificate.not_valid_after,
            subject_alternative_name=_SubjAltNameAsJson(dns=extract_dns_subject_alternative_names(certificate)),
            signature_hash_algorithm=signature_hash_algorithm,
            signature_algorithm_oid=certificate.signature_algorithm_oid,
            subject=subject_field,
            issuer=issuer_field,
            public_key=_PublicKeyAsJson.from_orm(certificate.public_key()),
        )


class _OcspResponseAsJson(_BaseModelWithOrmMode):
    response_status: str

    certificate_status: Optional[str]
    revocation_time: Optional[datetime]

    produced_at: Optional[datetime]
    this_update: Optional[datetime]
    next_update: Optional[datetime]

    serial_number: Optional[int]

    @classmethod
    def from_orm(cls, ocsp_response: x509.ocsp.OCSPResponse) -> "_OcspResponseAsJson":
        response_status = ocsp_response.response_status.name
        if ocsp_response.response_status != OCSPResponseStatus.SUCCESSFUL:
            return cls(
                response_status=response_status,
                certificate_status=None,
                revocation_time=None,
                produced_at=None,
                this_update=None,
                next_update=None,
                serial_number=None,
            )
        else:
            return cls(
                response_status=response_status,
                certificate_status=ocsp_response.certificate_status.name,
                revocation_time=ocsp_response.revocation_time,
                produced_at=ocsp_response.produced_at,
                this_update=ocsp_response.this_update,
                next_update=ocsp_response.next_update,
                serial_number=ocsp_response.serial_number,
            )


class _TrustStoreAsJson(_BaseModelWithOrmMode):
    path: Path
    name: str
    version: str
    ev_oids: Optional[List[_ObjectIdentifierAsJson]]


_TrustStoreAsJson.__doc__ = TrustStore.__doc__  # type: ignore


class _PathValidationResultAsJson(_BaseModelWithOrmMode):
    trust_store: _TrustStoreAsJson
    verified_certificate_chain: Optional[List[_CertificateAsJson]]
    openssl_error_string: Optional[str]
    was_validation_successful: bool


_PathValidationResultAsJson.__doc__ = PathValidationResult.__doc__  # type: ignore


class _CertificateDeploymentAnalysisResultAsJson(_BaseModelWithOrmMode):
    received_certificate_chain: List[_CertificateAsJson]
    leaf_certificate_subject_matches_hostname: bool
    leaf_certificate_has_must_staple_extension: bool
    leaf_certificate_is_ev: bool
    leaf_certificate_signed_certificate_timestamps_count: Optional[int]
    received_chain_contains_anchor_certificate: Optional[bool]
    received_chain_has_valid_order: Optional[bool]

    path_validation_results: List[_PathValidationResultAsJson]
    verified_chain_has_sha1_signature: Optional[bool]
    verified_chain_has_legacy_symantec_anchor: Optional[bool]

    ocsp_response: Optional[_OcspResponseAsJson]
    ocsp_response_is_trusted: Optional[bool]

    verified_certificate_chain: Optional[List[_CertificateAsJson]]


_CertificateDeploymentAnalysisResultAsJson.__doc__ = CertificateDeploymentAnalysisResult.__doc__  # type: ignore


class CertificateInfoScanResultAsJson(_BaseModelWithOrmMode):
    hostname_used_for_server_name_indication: str
    certificate_deployments: List[_CertificateDeploymentAnalysisResultAsJson]


CertificateInfoScanResultAsJson.__doc__ = CertificateInfoScanResult.__doc__  # type: ignore


class CertificateInfoScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[CertificateInfoScanResultAsJson]



from dataclasses import dataclass
from enum import Enum
from traceback import TracebackException
from typing import Generic, Optional, TypeVar


# These are in a separate file to avoid cyclic imports
_ScanCommandResultTypeVar = TypeVar("_ScanCommandResultTypeVar")


class ScanCommandAttemptStatusEnum(str, Enum):
    ERROR = "ERROR"
    COMPLETED = "COMPLETED"
    NOT_SCHEDULED = "NOT_SCHEDULED"


class ScanCommandErrorReasonEnum(str, Enum):
    BUG_IN_SSLYZE = "BUG_IN_SSLYZE"
    CLIENT_CERTIFICATE_NEEDED = "CLIENT_CERTIFICATE_NEEDED"
    CONNECTIVITY_ISSUE = "CONNECTIVITY_ISSUE"
    WRONG_USAGE = "WRONG_USAGE"


@dataclass(frozen=True)
class ScanCommandAttempt(Generic[_ScanCommandResultTypeVar]):
    """The result of a single scan command.

    Attributes:
        status: Whether this specific scan command was ran successfully.
        error_reason: The reason why the scan command failed; None if the scan command succeeded.
        error_trace: The exception trace of when the scan command failed; None if the scan command succeeded.
        result: The actual result of the scan command; None if the scan command failed. The type of this attribute is
            the "ScanResult" object corresponding to the scan command.
    """

    status: ScanCommandAttemptStatusEnum

    # Set if status == ERROR
    error_reason: Optional[ScanCommandErrorReasonEnum]
    error_trace: Optional[TracebackException]

    # Set if status == COMPLETED
    result: Optional[_ScanCommandResultTypeVar]



import socket
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple

import pydantic
from nassl._nassl import OpenSSLError
from nassl.legacy_ssl_client import LegacySslClient

from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.errors import ServerRejectedTlsHandshake
from sslyze.plugins.plugin_base import (
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandResult,
    ScanCommandWrongUsageError,
    ScanCommandCliConnector,
    ScanJobResult,
)
from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum


@dataclass(frozen=True)
class SessionRenegotiationScanResult(ScanCommandResult):
    """The result of testing a server for insecure TLS renegotiation and client-initiated renegotiation.

    Attributes:
        accepts_client_renegotiation: True if the server honors client-initiated renegotiation attempts.
        supports_secure_renegotiation: True if the server supports secure renegotiation.
    """

    supports_secure_renegotiation: bool
    is_vulnerable_to_client_renegotiation_dos: bool


# Identical fields in the JSON output
SessionRenegotiationScanResultAsJson = pydantic.dataclasses.dataclass(SessionRenegotiationScanResult, frozen=True)


class SessionRenegotiationScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[SessionRenegotiationScanResultAsJson]  # type: ignore


class _ScanJobResultEnum(Enum):
    IS_VULNERABLE_TO_CLIENT_RENEG_DOS = 1
    SUPPORTS_SECURE_RENEG = 2


class _SessionRenegotiationCliConnector(ScanCommandCliConnector[SessionRenegotiationScanResult, None]):

    _cli_option = "reneg"
    _cli_description = "Test a server for for insecure TLS renegotiation and client-initiated renegotiation."

    @classmethod
    def result_to_console_output(cls, result: SessionRenegotiationScanResult) -> List[str]:
        result_txt = [cls._format_title("Session Renegotiation")]

        # Client-initiated reneg
        client_reneg_txt = (
            "VULNERABLE - Server honors client-initiated renegotiations"
            if result.is_vulnerable_to_client_renegotiation_dos
            else "OK - Not vulnerable"
        )
        result_txt.append(cls._format_field("Client Renegotiation DoS Attack:", client_reneg_txt))

        # Secure reneg
        secure_txt = (
            "OK - Supported"
            if result.supports_secure_renegotiation
            else "VULNERABLE - Secure renegotiation not supported"
        )
        result_txt.append(cls._format_field("Secure Renegotiation:", secure_txt))

        return result_txt


class SessionRenegotiationImplementation(ScanCommandImplementation[SessionRenegotiationScanResult, None]):
    """Test a server for insecure TLS renegotiation and client-initiated renegotiation.
    """

    cli_connector_cls = _SessionRenegotiationCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        return [
            ScanJob(function_to_call=_test_secure_renegotiation, function_arguments=[server_info]),
            ScanJob(function_to_call=_test_client_renegotiation, function_arguments=[server_info]),
        ]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> SessionRenegotiationScanResult:
        if len(scan_job_results) != 2:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        results_dict = {}
        for job in scan_job_results:
            result_enum, value = job.get_result()
            results_dict[result_enum] = value

        return SessionRenegotiationScanResult(
            is_vulnerable_to_client_renegotiation_dos=results_dict[
                _ScanJobResultEnum.IS_VULNERABLE_TO_CLIENT_RENEG_DOS
            ],
            supports_secure_renegotiation=results_dict[_ScanJobResultEnum.SUPPORTS_SECURE_RENEG],
        )


def _test_secure_renegotiation(server_info: ServerConnectivityInfo) -> Tuple[_ScanJobResultEnum, bool]:
    """Check whether the server supports secure renegotiation.
    """
    # Try with TLS 1.2 even if the server supports TLS 1.3 or higher as there is no reneg with TLS 1.3
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        tls_version_to_use = TlsVersionEnum.TLS_1_2
        downgraded_from_tls_1_3 = True
    else:
        tls_version_to_use = server_info.tls_probing_result.highest_tls_version_supported
        downgraded_from_tls_1_3 = False

    ssl_connection = server_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version_to_use,
        should_use_legacy_openssl=True,  # Only the legacy SSL client has methods to check for secure reneg
    )
    if not isinstance(ssl_connection.ssl_client, LegacySslClient):
        raise RuntimeError("Should never happen")

    try:
        # Perform the TLS handshake
        ssl_connection.connect()
        supports_secure_renegotiation = ssl_connection.ssl_client.get_secure_renegotiation_support()

    # Should only happen when the server only supports TLS 1.3
    except ServerRejectedTlsHandshake:
        if downgraded_from_tls_1_3:
            supports_secure_renegotiation = True  # Technically TLS 1.3 has no renegotiation therefore it is secure
        else:
            raise

    finally:
        ssl_connection.close()

    return _ScanJobResultEnum.SUPPORTS_SECURE_RENEG, supports_secure_renegotiation


def _test_client_renegotiation(server_info: ServerConnectivityInfo) -> Tuple[_ScanJobResultEnum, bool]:
    """Check whether the server honors session renegotiation requests.
    """
    # Try with TLS 1.2 even if the server supports TLS 1.3 or higher as there is no reneg with TLS 1.3
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        tls_version_to_use = TlsVersionEnum.TLS_1_2
        downgraded_from_tls_1_3 = True
    else:
        tls_version_to_use = server_info.tls_probing_result.highest_tls_version_supported
        downgraded_from_tls_1_3 = False

    ssl_connection = server_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version_to_use,
        should_use_legacy_openssl=True,  # Only the legacy SSL client has methods to trigger a reneg
    )
    if not isinstance(ssl_connection.ssl_client, LegacySslClient):
        raise RuntimeError("Should never happen")

    try:
        # Perform the TLS handshake
        ssl_connection.connect()

    # Should only happen when the server only supports TLS 1.3
    except ServerRejectedTlsHandshake:
        if downgraded_from_tls_1_3:
            accepts_client_renegotiation = False  # Technically TLS 1.3 has no renegotiation therefore it is secure
        else:
            raise

    # The initial TLS handshake went well; let's try to do a renegotiation
    else:
        try:
            # Do a reneg multiple times in a row to be 100% sure that the server has no mitigations in place
            # https://github.com/nabla-c0d3/sslyze/issues/473
            for i in range(10):
                ssl_connection.ssl_client.do_renegotiate()
            accepts_client_renegotiation = True

        # Errors caused by a server rejecting the renegotiation
        except socket.timeout:
            # This is how Netty rejects a renegotiation - https://github.com/nabla-c0d3/sslyze/issues/114
            accepts_client_renegotiation = False
        except ConnectionError:
            accepts_client_renegotiation = False
        except OSError as e:
            # OSError is the parent of all (non-TLS) socket/connection errors so it should be last
            if "Nassl SSL handshake failed" in e.args[0]:
                # Special error returned by nassl
                accepts_client_renegotiation = False
            else:
                raise
        except OpenSSLError as e:
            if "handshake failure" in e.args[0]:
                accepts_client_renegotiation = False
            elif "no renegotiation" in e.args[0]:
                accepts_client_renegotiation = False
            elif "tlsv1 unrecognized name" in e.args[0]:
                # Yahoo's very own way of rejecting a renegotiation
                accepts_client_renegotiation = False
            elif "tlsv1 alert internal error" in e.args[0]:
                # Jetty server: https://github.com/nabla-c0d3/sslyze/issues/290
                accepts_client_renegotiation = False
            elif "decryption failed or bad record mac" in e.args[0]:
                # Some servers such as reddit.com
                accepts_client_renegotiation = False
            elif "sslv3 alert unexpected message" in e.args[0]:
                # traefik https://github.com/nabla-c0d3/sslyze/issues/422
                accepts_client_renegotiation = False
            elif "shut down by peer" in e.args[0]:
                # Cloudfront
                accepts_client_renegotiation = False
            elif "unexpected record" in e.args[0]:
                # Indy TCP Server with special RSA Token authentication https://github.com/nabla-c0d3/sslyze/issues/483
                accepts_client_renegotiation = False

            else:
                raise

    finally:
        ssl_connection.close()

    return _ScanJobResultEnum.IS_VULNERABLE_TO_CLIENT_RENEG_DOS, accepts_client_renegotiation



from sslyze import __version__


class HttpRequestGenerator:

    HTTP_GET_FORMAT = (
        "GET {path} HTTP/1.1\r\n"
        "Host: {host}\r\n"
        "User-Agent: {user_agent}\r\n"
        "Accept: */*\r\n"
        "Connection: close\r\n\r\n"
    )

    DEFAULT_USER_AGENT = (
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/52.0.2743.116 Safari/537.36 SSLyze/{0}".format(__version__.__version__)
    )

    @classmethod
    def get_request(cls, host: str, path: str = "/") -> bytes:
        return cls.HTTP_GET_FORMAT.format(host=host, path=path, user_agent=cls.DEFAULT_USER_AGENT).encode("utf-8")



# flake8: noqa

# Classes for configuring the servers to scan
from sslyze.server_setting import (
    ServerNetworkLocation,
    ConnectionTypeEnum,
    HttpProxySettings,
    ClientAuthenticationCredentials,
    ServerNetworkConfiguration,
)
from sslyze.connection_helpers.opportunistic_tls_helpers import ProtocolWithOpportunisticTlsEnum
from nassl.ssl_client import OpenSslFileTypeEnum


# Classes for testing connectivity with the servers
from sslyze.server_connectivity import (
    ClientAuthRequirementEnum,
    TlsVersionEnum,
    ServerTlsProbingResult,
)

# Classes for setting up scan commands and extra arguments
from sslyze.plugins.scan_commands import ScanCommand
from sslyze.plugins.certificate_info.implementation import CertificateInfoExtraArgument

# Classes for scanning the servers
from sslyze.scanner.models import (
    ServerScanRequest,
    ScanCommandsExtraArguments,
    ServerScanResult,
    ServerConnectivityStatusEnum,
    ServerScanStatusEnum,
    AllScanCommandsAttempts,
)
from sslyze.scanner.scan_command_attempt import (
    ScanCommandAttempt,
    ScanCommandAttemptStatusEnum,
    ScanCommandErrorReasonEnum,
)
from sslyze.scanner.scanner import Scanner
from sslyze.errors import ServerHostnameCouldNotBeResolved


# Classes with the scan results
from sslyze.plugins.plugin_base import ScanCommandResult

# Certificate Info
from sslyze.plugins.certificate_info.implementation import (
    CertificateInfoScanResult,
    CertificateDeploymentAnalysisResult,
)
from sslyze.plugins.certificate_info._cert_chain_analyzer import PathValidationResult
from sslyze.plugins.certificate_info.trust_stores.trust_store import TrustStore

# Cipher Suites
from sslyze.plugins.openssl_cipher_suites.implementation import CipherSuitesScanResult
from sslyze.plugins.openssl_cipher_suites._test_cipher_suite import (
    CipherSuiteAcceptedByServer,
    CipherSuiteRejectedByServer,
)
from nassl.ephemeral_key_info import EphemeralKeyInfo
from sslyze.plugins.openssl_cipher_suites.cipher_suites import CipherSuite

from sslyze.plugins.robot.implementation import RobotScanResult, RobotScanResultEnum

from sslyze.plugins.session_resumption.implementation import (
    SessionResumptionSupportScanResult,
    TlsResumptionSupportEnum,
    SessionResumptionSupportExtraArgument,
)
from sslyze.plugins.compression_plugin import CompressionScanResult
from sslyze.plugins.early_data_plugin import EarlyDataScanResult
from sslyze.plugins.fallback_scsv_plugin import FallbackScsvScanResult
from sslyze.plugins.heartbleed_plugin import HeartbleedScanResult

# HTTP Headers
from sslyze.plugins.http_headers_plugin import (
    HttpHeadersScanResult,
    StrictTransportSecurityHeader,
    ExpectCtHeader,
)


from sslyze.plugins.openssl_ccs_injection_plugin import OpenSslCcsInjectionScanResult
from sslyze.plugins.session_renegotiation_plugin import SessionRenegotiationScanResult
from sslyze.plugins.elliptic_curves_plugin import SupportedEllipticCurvesScanResult, EllipticCurve

from sslyze.json.json_output import SslyzeOutputAsJson, ServerScanResultAsJson



import socket
from dataclasses import dataclass
from typing import Tuple, Optional


@dataclass(frozen=True)
class InvalidServerStringError(Exception):
    """Exception raised when SSLyze was unable to parse a hostname:port string supplied via the command line.
    """

    server_string: str
    error_message: str


class CommandLineServerStringParser:
    """Utility class to parse a 'host:port{ip}' string taken from the command line into a valid (host,ip, port) tuple.
    Supports IPV6 addresses.
    """

    SERVER_STRING_ERROR_BAD_PORT = "Not a valid host:port"

    @classmethod
    def parse_server_string(cls, server_str: str) -> Tuple[str, Optional[str], Optional[int]]:
        # Extract ip from target
        ip = None
        if "{" in server_str and "}" in server_str:
            raw_target = server_str.split("{")
            raw_ip = raw_target[1]

            ip = raw_ip.replace("}", "")

            # Clean the target
            server_str = raw_target[0]

        # Look for ipv6 hint in target
        if "[" in server_str:
            (host, port) = cls._parse_ipv6_server_string(server_str)
        else:
            # Look for ipv6 hint in the ip
            if ip is not None and "[" in ip:
                (ip, port) = cls._parse_ipv6_server_string(ip)

            # Fallback to ipv4
            (host, port) = cls._parse_ipv4_server_string(server_str)

        return host, ip, port

    @classmethod
    def _parse_ipv4_server_string(cls, server_str: str) -> Tuple[str, Optional[int]]:
        host = server_str
        port = None
        if ":" in server_str:
            host = (server_str.split(":"))[0]  # hostname or ipv4 address
            try:
                port = int((server_str.split(":"))[1])
            except ValueError:  # Port is not an int
                raise InvalidServerStringError(server_string=server_str, error_message=cls.SERVER_STRING_ERROR_BAD_PORT)

        return host, port

    @classmethod
    def _parse_ipv6_server_string(cls, server_str: str) -> Tuple[str, Optional[int]]:
        if not socket.has_ipv6:
            raise InvalidServerStringError(
                server_string=server_str, error_message="IPv6 is not supported on this platform"
            )

        port = None
        target_split = server_str.split("]")
        ipv6_addr = target_split[0].split("[")[1]
        if ":" in target_split[1]:  # port was specified
            try:
                port = int(target_split[1].rsplit(":")[1])
            except ValueError:  # Port is not an int
                raise InvalidServerStringError(server_string=server_str, error_message=cls.SERVER_STRING_ERROR_BAD_PORT)
        return ipv6_addr, port



import socket
import types
from dataclasses import dataclass
from typing import List, Optional

import pydantic
from nassl._nassl import WantReadError

from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.plugins.plugin_base import (
    ScanCommandResult,
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandWrongUsageError,
    ScanCommandCliConnector,
    ScanJobResult,
)
from tls_parser.alert_protocol import TlsAlertRecord
from tls_parser.application_data_protocol import TlsApplicationDataRecord
from tls_parser.change_cipher_spec_protocol import TlsChangeCipherSpecRecord
from tls_parser.exceptions import NotEnoughData, UnknownTlsVersionByte
from tls_parser.handshake_protocol import TlsHandshakeRecord, TlsHandshakeTypeByte
from tls_parser.parser import TlsRecordParser
import tls_parser.tls_version

from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum


@dataclass(frozen=True)
class OpenSslCcsInjectionScanResult(ScanCommandResult):
    """The result of testing a server for the OpenSSL CCS Injection vulnerability (CVE-2014-0224).

    Attributes:
        is_vulnerable_to_ccs_injection: True if the server is vulnerable to the OpenSSL CCS Injection vulnerability.
    """

    is_vulnerable_to_ccs_injection: bool


# Identical fields in the JSON output
OpenSslCcsInjectionScanResultAsJson = pydantic.dataclasses.dataclass(OpenSslCcsInjectionScanResult, frozen=True)


class OpenSslCcsInjectionScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[OpenSslCcsInjectionScanResultAsJson]  # type: ignore


class _OpenSslCcsInjectionCliConnector(ScanCommandCliConnector[OpenSslCcsInjectionScanResult, None]):

    _cli_option = "openssl_ccs"
    _cli_description = "Test a server for the OpenSSL CCS Injection vulnerability (CVE-2014-0224)."

    @classmethod
    def result_to_console_output(cls, result: OpenSslCcsInjectionScanResult) -> List[str]:
        result_txt = [cls._format_title("OpenSSL CCS Injection")]
        ccs_text = (
            "VULNERABLE - Server is vulnerable to OpenSSL CCS injection"
            if result.is_vulnerable_to_ccs_injection
            else "OK - Not vulnerable to OpenSSL CCS injection"
        )
        result_txt.append(cls._format_field("", ccs_text))
        return result_txt


class OpenSslCcsInjectionImplementation(ScanCommandImplementation[OpenSslCcsInjectionScanResult, None]):
    """Test a server for the OpenSSL CCS Injection vulnerability (CVE-2014-0224).
    """

    cli_connector_cls = _OpenSslCcsInjectionCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        return [ScanJob(function_to_call=_test_for_ccs_injection, function_arguments=[server_info])]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> OpenSslCcsInjectionScanResult:
        if len(scan_job_results) != 1:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        return OpenSslCcsInjectionScanResult(is_vulnerable_to_ccs_injection=scan_job_results[0].get_result())


def _test_for_ccs_injection(server_info: ServerConnectivityInfo) -> bool:
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        # The server uses a recent version of OpenSSL and it cannot be vulnerable to CCS Injection
        return False

    # Disable SNI for this check because some legacy servers don't support sending the CCS payload and SNI
    # See https://github.com/nabla-c0d3/sslyze/issues/202
    ssl_connection = server_info.get_preconfigured_tls_connection(should_enable_server_name_indication=False)

    # Replace nassl.sslClient.do_handshake() with a CCS checking SSL handshake so that all the SSLyze options
    # (startTLS, proxy, etc.) still work
    ssl_connection.ssl_client.do_handshake = types.MethodType(  # type: ignore
        _do_handshake_with_ccs_injection, ssl_connection.ssl_client
    )

    is_vulnerable = False
    try:
        # Start the SSL handshake
        ssl_connection.connect()
    except _VulnerableToCcsInjection:
        # The test was completed and the server is vulnerable
        is_vulnerable = True
    except _NotVulnerableToCcsInjection:
        # The test was completed and the server is NOT vulnerable
        pass
    finally:
        ssl_connection.close()

    return is_vulnerable


class _VulnerableToCcsInjection(Exception):
    """Exception to raise during the handshake to hijack the flow and test for CCS.
    """


class _NotVulnerableToCcsInjection(Exception):
    """Exception to raise during the handshake to hijack the flow and test for CCS.
    """


def _do_handshake_with_ccs_injection(self):  # type: ignore
    """Modified do_handshake() to send a CCS injection payload and return the result.
    """
    try:
        # Start the handshake using nassl - will throw WantReadError right away
        self._ssl.do_handshake()
    except WantReadError:
        # Send the Client Hello
        len_to_read = self._network_bio.pending()
        while len_to_read:
            # Get the data from the SSL engine
            handshake_data_out = self._network_bio.read(len_to_read)
            # Send it to the peer
            self._sock.send(handshake_data_out)
            len_to_read = self._network_bio.pending()

    # Retrieve the server's response - directly read the underlying network socket
    # Retrieve data until we get to the ServerHelloDone
    # The server may send back a ServerHello, an Alert or a CertificateRequest first
    did_receive_hello_done = False
    remaining_bytes = b""
    while not did_receive_hello_done:
        try:
            tls_record, len_consumed = TlsRecordParser.parse_bytes(remaining_bytes)
            remaining_bytes = remaining_bytes[len_consumed::]
        except UnknownTlsVersionByte as e:
            # Workaround for Amazon Cloudfront; see https://github.com/nabla-c0d3/sslyze/issues/437
            if e.record_type == tls_parser.record_protocol.TlsRecordTypeByte.ALERT:
                # Server returned a (badly-formatted) TLS alert because it requires SNI
                # Hence the server uses a modern TLS stack and is not vulnerable
                raise _NotVulnerableToCcsInjection()
            else:
                raise
        except NotEnoughData:
            # Try to get more data
            raw_ssl_bytes = self._sock.recv(16381)
            if not raw_ssl_bytes:
                # No data?
                break

            remaining_bytes = remaining_bytes + raw_ssl_bytes
            continue

        if isinstance(tls_record, TlsHandshakeRecord):
            # Does the record contain a ServerDone message?
            for handshake_message in tls_record.subprotocol_messages:
                if handshake_message.handshake_type == TlsHandshakeTypeByte.SERVER_DONE:
                    did_receive_hello_done = True
                    break
            # If not, it could be a ServerHello, Certificate or a CertificateRequest if the server requires client auth
        elif isinstance(tls_record, TlsAlertRecord):
            # Server returned a TLS alert
            break
        else:
            raise ValueError("Unknown record? Type {}".format(tls_record.header.type))

    if did_receive_hello_done:
        # Send an early CCS record - this should be rejected by the server
        payload = TlsChangeCipherSpecRecord.from_parameters(
            tls_version=tls_parser.tls_version.TlsVersionEnum[self._ssl_version.name]
        ).to_bytes()
        self._sock.send(payload)

        # Send an early application data record which should be ignored by the server
        app_data_record = TlsApplicationDataRecord.from_parameters(
            tls_version=tls_parser.tls_version.TlsVersionEnum[self._ssl_version.name], application_data=b"\x00\x00"
        )
        self._sock.send(app_data_record.to_bytes())

        # Check if an alert was sent back
        while True:
            try:
                tls_record, len_consumed = TlsRecordParser.parse_bytes(remaining_bytes)
                remaining_bytes = remaining_bytes[len_consumed::]
            except NotEnoughData:
                # Try to get more data
                try:
                    raw_ssl_bytes = self._sock.recv(16381)
                    if not raw_ssl_bytes:
                        # No data?
                        raise _NotVulnerableToCcsInjection()
                except socket.error:
                    # Server closed the connection after receiving the CCS payload
                    raise _NotVulnerableToCcsInjection()

                remaining_bytes = remaining_bytes + raw_ssl_bytes
                continue

            if isinstance(tls_record, TlsAlertRecord):
                # Server returned a TLS alert but which one?
                if tls_record.alert_description == 0x14:
                    # BAD_RECORD_MAC: This means that the server actually tried to decrypt our early application data
                    # record instead of ignoring it; server is vulnerable
                    raise _VulnerableToCcsInjection()

                # Any other alert means that the server rejected the early CCS record
                raise _NotVulnerableToCcsInjection()
            else:
                break

        raise _NotVulnerableToCcsInjection()



from dataclasses import dataclass
from argparse import ArgumentParser
from pathlib import Path

from nassl.ssl_client import OpenSslFileTypeEnum
from typing import Set, List, Optional, Dict
from typing import Tuple

from sslyze.cli.server_string_parser import (
    InvalidServerStringError,
    CommandLineServerStringParser,
)
from sslyze.connection_helpers.opportunistic_tls_helpers import ProtocolWithOpportunisticTlsEnum
from sslyze.mozilla_tls_profile.mozilla_config_checker import (
    MozillaTlsConfigurationEnum,
    SCAN_COMMANDS_NEEDED_BY_MOZILLA_CHECKER,
)
from sslyze.plugins import plugin_base
from sslyze.plugins.certificate_info.trust_stores.trust_store_repository import TrustStoresRepository
from sslyze.plugins.plugin_base import OptParseCliOption
from sslyze.plugins.scan_commands import ScanCommand, ScanCommandsRepository
from sslyze.scanner.models import ScanCommandsExtraArguments

from sslyze.server_setting import (
    HttpProxySettings,
    ServerNetworkLocation,
    ServerNetworkConfiguration,
    InvalidServerNetworkConfigurationError,
    ServerHostnameCouldNotBeResolved,
    ClientAuthenticationCredentials,
)


class CommandLineParsingError(Exception):

    PARSING_ERROR_FORMAT = "  Command line error: {0}\n  Use -h for help."

    def get_error_msg(self) -> str:
        return self.PARSING_ERROR_FORMAT.format(self)


# Not actually an error but makes handling trust store updates a lot easier
class TrustStoresUpdateCompleted(CommandLineParsingError):
    def get_error_msg(self) -> str:
        return "Trust stores successfully updated."


@dataclass(frozen=True)
class ParsedCommandLine:
    """The result of parsing a command line used to launch sslyze.
    """

    invalid_servers: List[InvalidServerStringError]

    # Servers to scan
    servers_to_scans: List[Tuple[ServerNetworkLocation, ServerNetworkConfiguration]]
    scan_commands: Set[ScanCommand]
    scan_commands_extra_arguments: ScanCommandsExtraArguments

    # Output settings
    json_path_out: Optional[Path]
    should_print_json_to_console: bool
    should_disable_console_output: bool

    # Network settings
    per_server_concurrent_connections_limit: Optional[int]
    concurrent_server_scans_limit: Optional[int]

    # Mozilla compliance; None if shouldn't be run
    check_against_mozilla_config: Optional[MozillaTlsConfigurationEnum]


_STARTTLS_PROTOCOL_DICT = {
    "smtp": ProtocolWithOpportunisticTlsEnum.SMTP,
    "xmpp": ProtocolWithOpportunisticTlsEnum.XMPP,
    "xmpp_server": ProtocolWithOpportunisticTlsEnum.XMPP_SERVER,
    "pop3": ProtocolWithOpportunisticTlsEnum.POP3,
    "imap": ProtocolWithOpportunisticTlsEnum.IMAP,
    "ftp": ProtocolWithOpportunisticTlsEnum.FTP,
    "ldap": ProtocolWithOpportunisticTlsEnum.LDAP,
    "rdp": ProtocolWithOpportunisticTlsEnum.RDP,
    "postgres": ProtocolWithOpportunisticTlsEnum.POSTGRES,
}


class CommandLineParser:
    def __init__(self, sslyze_version: str) -> None:
        """Generate SSLyze's command line parser.
        """
        self.aparser = ArgumentParser(prog="sslyze", description=f"SSLyze version {sslyze_version}")

        # Add generic command line options to the parser
        self._add_default_options()

        # Add plugin .ie scan command options to the parser
        scan_commands_group = self.aparser.add_argument_group("Scan commands")
        for scan_option in self._get_plugin_scan_commands():
            scan_commands_group.add_argument(
                f"--{scan_option.option}", help=scan_option.help, action=scan_option.action,
            )

        self.aparser.add_argument(
            "--mozilla-config",
            action="store",
            dest="mozilla_config",
            choices=[config.value for config in MozillaTlsConfigurationEnum],
            help="Shortcut to queue various scan commands needed to check the server's TLS configurations against one"
            ' of Mozilla\'s recommended TLS configuration. Set to "intermediate" by default.',
        )

        self.aparser.add_argument(dest="target", default=[], nargs="*", help="The list of servers to scan.")

    def parse_command_line(self) -> ParsedCommandLine:
        """Parses the command line used to launch SSLyze.
        """
        args_command_list = self.aparser.parse_args()
        args_target_list = []

        if args_command_list.update_trust_stores:
            # Just update the trust stores and do nothing
            TrustStoresRepository.update_default()
            raise TrustStoresUpdateCompleted()

        # Handle the --targets_in command line and fill args_target_list
        if args_command_list.targets_in:
            try:  # Read targets from a file
                with open(args_command_list.targets_in) as f:
                    for target in f.readlines():
                        if target.strip():  # Ignore empty lines
                            if not target.startswith("#"):  # Ignore comment lines
                                args_target_list.append(target.strip())
            except IOError:
                raise CommandLineParsingError(f"Can't read targets from input file '{args_command_list.targets_in}'")

        for target in args_command_list.target:
            args_target_list.append(target)

        if not args_target_list:
            raise CommandLineParsingError("No targets to scan.")

        # Handle the case when no scan commands have been specified: run --mozilla-config=intermediate by default
        if not args_command_list.mozilla_config:
            did_user_enable_some_scan_commands = [
                getattr(args_command_list, option.option)
                for option in self._get_plugin_scan_commands()
                if getattr(args_command_list, option.option)
            ]
            if not did_user_enable_some_scan_commands:
                setattr(args_command_list, "mozilla_config", MozillaTlsConfigurationEnum.INTERMEDIATE.value)

        # Enable the commands needed by --mozilla-config
        check_against_mozilla_config: Optional[MozillaTlsConfigurationEnum] = None
        if args_command_list.mozilla_config:
            check_against_mozilla_config = MozillaTlsConfigurationEnum(args_command_list.mozilla_config)
            for scan_cmd in SCAN_COMMANDS_NEEDED_BY_MOZILLA_CHECKER:
                cli_connector_cls = ScanCommandsRepository.get_implementation_cls(scan_cmd).cli_connector_cls
                setattr(args_command_list, cli_connector_cls._cli_option, True)

        # Handle JSON settings
        should_print_json_to_console = False
        json_path_out: Optional[Path] = None
        if args_command_list.json_file:
            if args_command_list.json_file == "-":
                if args_command_list.quiet:
                    raise CommandLineParsingError("Cannot use --quiet with --json_out -.")
                should_print_json_to_console = True
            else:
                json_path_out = Path(args_command_list.json_file).absolute()

        # Sanity checks on the client cert options
        client_auth_creds = None
        if bool(args_command_list.cert) ^ bool(args_command_list.key):
            raise CommandLineParsingError("No private key or certificate file were given. See --cert and --key.")

        elif args_command_list.cert:
            # Private key formats
            if args_command_list.keyform == "DER":
                key_type = OpenSslFileTypeEnum.ASN1
            elif args_command_list.keyform == "PEM":
                key_type = OpenSslFileTypeEnum.PEM

            # Let's try to open the cert and key files
            try:
                client_auth_creds = ClientAuthenticationCredentials(
                    certificate_chain_path=Path(args_command_list.cert),
                    key_path=Path(args_command_list.key),
                    key_password=args_command_list.keypass,
                    key_type=key_type,
                )
            except ValueError as e:
                raise CommandLineParsingError("Invalid client authentication settings: {}.".format(e.args[0]))

        # HTTP CONNECT proxy
        http_proxy_settings = None
        if args_command_list.https_tunnel:
            try:
                http_proxy_settings = HttpProxySettings.from_url(args_command_list.https_tunnel)
            except ValueError as e:
                raise CommandLineParsingError("Invalid proxy URL for --https_tunnel: {}.".format(e.args[0]))

        # Create the server location objects for each specified servers
        good_servers: List[Tuple[ServerNetworkLocation, ServerNetworkConfiguration]] = []
        invalid_server_strings: List[InvalidServerStringError] = []
        for server_string in args_target_list:
            try:
                # Parse the string supplied via the CLI for this server
                (hostname, ip_address, port,) = CommandLineServerStringParser.parse_server_string(server_string)
            except InvalidServerStringError as e:
                # The server string is malformed
                invalid_server_strings.append(e)
                continue

            # If not port number was supplied, assume 443
            final_port = port if port else 443

            # Figure out how we're going to connect to the server
            server_location: ServerNetworkLocation
            if http_proxy_settings:
                # Connect to the server via an HTTP proxy
                # A limitation when using the CLI is that only one http_proxy_settings can be specified for all servers
                server_location = ServerNetworkLocation(
                    hostname=hostname, port=final_port, http_proxy_settings=http_proxy_settings,
                )
            else:
                # Connect to the server directly
                if ip_address:
                    server_location = ServerNetworkLocation(hostname=hostname, port=final_port, ip_address=ip_address,)
                else:
                    # No IP address supplied - do a DNS lookup
                    try:
                        server_location = ServerNetworkLocation(hostname=hostname, port=final_port,)
                    except ServerHostnameCouldNotBeResolved:
                        invalid_server_strings.append(
                            InvalidServerStringError(
                                server_string=f"{hostname}:{final_port}",
                                error_message=f"Could not resolve hostname {hostname}",
                            )
                        )
                        continue

            # Figure out extra network config for this server
            # Opportunistic TLS
            opportunistic_tls: Optional[ProtocolWithOpportunisticTlsEnum] = None
            if args_command_list.starttls:
                if args_command_list.starttls == "auto":
                    # Special value to auto-derive the protocol from the port number
                    opportunistic_tls = ProtocolWithOpportunisticTlsEnum.from_default_port(final_port)
                elif args_command_list.starttls in _STARTTLS_PROTOCOL_DICT:
                    opportunistic_tls = _STARTTLS_PROTOCOL_DICT[args_command_list.starttls]
                else:
                    raise CommandLineParsingError(
                        f"StartTLS should be one of: auto, {', '.join(_STARTTLS_PROTOCOL_DICT.keys())}."
                    )

            try:
                sni_hostname = args_command_list.sni if args_command_list.sni else hostname
                network_config = ServerNetworkConfiguration(
                    tls_opportunistic_encryption=opportunistic_tls,
                    tls_server_name_indication=sni_hostname,
                    tls_client_auth_credentials=client_auth_creds,
                    xmpp_to_hostname=args_command_list.xmpp_to,
                )
                good_servers.append((server_location, network_config))
            except InvalidServerNetworkConfigurationError as e:
                raise CommandLineParsingError(e.args[0])

        # Figure out global network settings
        concurrent_server_scans_limit = None
        per_server_concurrent_connections_limit = None
        if args_command_list.https_tunnel:
            # All the connections will go through a single proxy; only scan one server at a time to not DOS the proxy
            concurrent_server_scans_limit = 1
        if args_command_list.slow_connection:
            # Go easy on the servers; only open 2 concurrent connections against each server
            per_server_concurrent_connections_limit = 2

        # Figure out the scan commands that are enabled
        scan_commands: Set[ScanCommand] = set()
        scan_commands_extra_arguments_dict: Dict[ScanCommand, plugin_base.ScanCommandExtraArgument] = {}
        for scan_command in ScanCommandsRepository.get_all_scan_commands():
            cli_connector_cls = ScanCommandsRepository.get_implementation_cls(scan_command).cli_connector_cls
            (is_scan_cmd_enabled, extra_args,) = cli_connector_cls.find_cli_options_in_command_line(
                args_command_list.__dict__
            )
            if is_scan_cmd_enabled:
                scan_commands.add(scan_command)
                if extra_args:
                    scan_commands_extra_arguments_dict[scan_command] = extra_args
        scan_commands_extra_arguments = ScanCommandsExtraArguments(**scan_commands_extra_arguments_dict)  # type: ignore

        return ParsedCommandLine(
            invalid_servers=invalid_server_strings,
            servers_to_scans=good_servers,
            scan_commands=scan_commands,
            scan_commands_extra_arguments=scan_commands_extra_arguments,
            should_print_json_to_console=should_print_json_to_console,
            json_path_out=json_path_out,
            should_disable_console_output=args_command_list.quiet or args_command_list.json_file == "-",
            concurrent_server_scans_limit=concurrent_server_scans_limit,
            per_server_concurrent_connections_limit=per_server_concurrent_connections_limit,
            check_against_mozilla_config=check_against_mozilla_config,
        )

    def _add_default_options(self) -> None:
        """Add default command line options to the parser.
        """
        # Updating the trust stores
        trust_stores_group = self.aparser.add_argument_group("Trust stores options")
        trust_stores_group.add_argument(
            "--update_trust_stores",
            help="Update the default trust stores used by SSLyze. The latest stores will be "
            "downloaded from https://github.com/nabla-c0d3/trust_stores_observatory. This option "
            "is meant to be used separately, and will silence any other command line option "
            "supplied to SSLyze.",
            dest="update_trust_stores",
            action="store_true",
        )

        # Client certificate options
        client_certificate_group = self.aparser.add_argument_group("Client certificate options")
        client_certificate_group.add_argument(
            "--cert",
            metavar="CERTIFICATE_FILE",
            help="Client certificate chain filename. The certificates must be in PEM format and "
            "must be sorted starting with the subject's client certificate, followed by "
            "intermediate CA certificates if applicable.",
            dest="cert",
        )
        client_certificate_group.add_argument(
            "--key", metavar="KEY_FILE", help="Client private key filename.", dest="key"
        )
        client_certificate_group.add_argument(
            "--keyform",
            metavar="KEY_FORMAT",
            choices=["DER", "PEM"],
            help="Client private key format. DER or PEM " "(default).",
            dest="keyform",
            default="PEM",
        )
        client_certificate_group.add_argument(
            "--pass", metavar="PASSPHRASE", help="Client private key passphrase.", dest="keypass", default="",
        )

        # Input / output
        input_and_output_group = self.aparser.add_argument_group("Input and output options")
        # JSON output
        input_and_output_group.add_argument(
            "--json_out",
            metavar="JSON_FILE",
            help="Write the scan results as a JSON document to the file JSON_FILE. If JSON_FILE "
            "is set to '-', the JSON output will instead be printed to stdout. The resulting "
            "JSON file is a serialized version of the ScanResult objects described in SSLyze's "
            "Python API: the nodes and attributes will be the same. "
            "See https://nabla-c0d3.github.io/sslyze/documentation/available-scan-commands.html "
            "for more details.",
            dest="json_file",
            default=None,
        )
        # Read targets from input file
        input_and_output_group.add_argument(
            "--targets_in",
            metavar="TARGET_FILE",
            help="Read the list of targets to scan from the file TARGET_FILE. It should contain "
            "one host:port per line.",
            dest="targets_in",
            default=None,
        )
        # No text output
        input_and_output_group.add_argument(
            "--quiet",
            action="store_true",
            dest="quiet",
            help="Do not output anything to stdout; useful when using --json_out.",
        )

        # Connectivity option group
        connectivity_group = self.aparser.add_argument_group("Contectivity options")
        # Connection speed
        connectivity_group.add_argument(
            "--slow_connection",
            help="Greatly reduce the number of concurrent connections initiated by SSLyze. This "
            "will make the scans slower but more reliable if the connection between your host and "
            "the server is slow, or if the server cannot handle many concurrent connections. "
            "Enable this option if you are getting a lot of timeouts or errors.",
            action="store_true",
            dest="slow_connection",
        )
        # HTTP CONNECT Proxy
        connectivity_group.add_argument(
            "--https_tunnel",
            metavar="PROXY_SETTINGS",
            help="Tunnel all traffic to the target server(s) through an HTTP CONNECT proxy. "
            "HTTP_TUNNEL should be the proxy's URL: 'http://USER:PW@HOST:PORT/'. For proxies "
            "requiring authentication, only Basic Authentication is supported.",
            dest="https_tunnel",
            default=None,
        )
        # STARTTLS
        connectivity_group.add_argument(
            "--starttls",
            metavar="PROTOCOL",
            help="Perform a StartTLS handshake when connecting to the target server(s). "
            f"StartTLS should be one of: auto, {', '.join(_STARTTLS_PROTOCOL_DICT.keys())}. The "
            "'auto' option will cause SSLyze to deduce the protocol (ftp, imap, etc.) from the "
            "supplied port number, for each target servers.",
            dest="starttls",
            default=None,
        )
        connectivity_group.add_argument(
            "--xmpp_to",
            metavar="HOSTNAME",
            help="Optional setting for STARTTLS XMPP. XMPP_TO should be the hostname to be put in "
            "the 'to' attribute of the XMPP stream. Default is the server's hostname.",
            dest="xmpp_to",
            default=None,
        )
        # Server Name Indication
        connectivity_group.add_argument(
            "--sni",
            metavar="SERVER_NAME_INDICATION",
            help="Use Server Name Indication to specify the hostname to connect to. Will only "
            "affect TLS 1.0+ connections.",
            dest="sni",
            default=None,
        )

    @staticmethod
    def _get_plugin_scan_commands() -> List[OptParseCliOption]:
        """Retrieve the list of command line options implemented by the plugins currently available.
        """
        scan_commands_options = []
        for scan_command in ScanCommandsRepository.get_all_scan_commands():
            cli_connector_cls = ScanCommandsRepository.get_implementation_cls(scan_command).cli_connector_cls
            scan_commands_options.extend(cli_connector_cls.get_cli_options())
        return scan_commands_options



from enum import Enum
from typing import Optional, Tuple

import nassl

from sslyze.errors import ServerRejectedTlsHandshake
from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum


class TlsResumptionSupportEnum(str, Enum):
    """The result of attempting to resume TLS sessions with the server.

    Attributes:
        FULLY_SUPPORTED: All the session resumption attempts were successful.
        PARTIALLY_SUPPORTED: Only some of the session resumption attempts were successful.
        NOT_SUPPORTED: None of the session resumption attempts were successful.
        SERVER_IS_TLS_1_3_ONLY: The server only supports TLS 1.3, which does not support Session ID nor TLS Tickets
            resumption.
    """

    FULLY_SUPPORTED = "FULLY_SUPPORTED"
    PARTIALLY_SUPPORTED = "PARTIALLY_SUPPORTED"
    NOT_SUPPORTED = "NOT_SUPPORTED"
    SERVER_IS_TLS_1_3_ONLY = "SERVER_IS_TLS_1_3_ONLY"


class _ScanJobResultEnum(Enum):
    TLS_TICKET_RESUMPTION = 1
    SESSION_ID_RESUMPTION = 2


class ServerOnlySupportsTls13(Exception):
    """If the server only supports TLS 1.3 or higher, it does not support session resumption with IDs or tickets.
    """

    pass


def retrieve_tls_session(
    server_info: ServerConnectivityInfo,
    session_to_resume: Optional[nassl._nassl.SSL_SESSION] = None,
    should_enable_tls_ticket: bool = False,
) -> nassl._nassl.SSL_SESSION:
    """Connect to the server and returns the session object that was assigned for that connection.

    If ssl_session is given, tries to resume that session.
    """
    # Try with TLS 1.2 even if the server supports TLS 1.3 or higher as there is no session resumption (with IDs or
    # tickets) with TLS 1.3
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        tls_version_to_use = TlsVersionEnum.TLS_1_2
        downgraded_from_tls_1_3 = True
    else:
        tls_version_to_use = server_info.tls_probing_result.highest_tls_version_supported
        downgraded_from_tls_1_3 = False

    ssl_connection = server_info.get_preconfigured_tls_connection(override_tls_version=tls_version_to_use)
    if not should_enable_tls_ticket:
        # Need to disable TLS tickets to test session IDs, according to rfc5077:
        # If a ticket is presented by the client, the server MUST NOT attempt
        # to use the Session ID in the ClientHello for stateful session resumption
        ssl_connection.ssl_client.disable_stateless_session_resumption()  # Turning off TLS tickets.

    if session_to_resume:
        ssl_connection.ssl_client.set_session(session_to_resume)

    try:
        # Perform the TLS handshake
        ssl_connection.connect()
        new_session = ssl_connection.ssl_client.get_session()  # Get session data

    except ServerRejectedTlsHandshake:
        if downgraded_from_tls_1_3:
            raise ServerOnlySupportsTls13()
        else:
            raise

    finally:
        ssl_connection.close()

    return new_session


def _extract_session_id(ssl_session: nassl._nassl.SSL_SESSION) -> str:
    """Extract the SSL session ID from a SSL session object or raises IndexError if the session ID was not set.
    """
    session_string = ((ssl_session.as_text()).split("Session-ID:"))[1]
    session_id = (session_string.split("Session-ID-ctx:"))[0].strip()
    return session_id


def resume_with_session_id(server_info: ServerConnectivityInfo) -> Tuple[_ScanJobResultEnum, bool]:
    """Perform one session resumption using Session IDs.
    """
    # Create a new TLS session with the server
    session1 = retrieve_tls_session(server_info)
    try:
        # Recover the session ID
        session1_id = _extract_session_id(session1)
    except IndexError:
        # Session ID not assigned
        return _ScanJobResultEnum.SESSION_ID_RESUMPTION, False

    if session1_id == "":
        # Session ID empty
        return _ScanJobResultEnum.SESSION_ID_RESUMPTION, False

    # Try to resume that TLS session
    session2 = retrieve_tls_session(server_info, session_to_resume=session1)
    try:
        # Recover the session ID
        session2_id = _extract_session_id(session2)
    except IndexError:
        # Session ID not assigned
        return _ScanJobResultEnum.SESSION_ID_RESUMPTION, False

    # Finally, compare the two Session IDs
    if session1_id != session2_id:
        # Session ID assigned but not accepted
        return _ScanJobResultEnum.SESSION_ID_RESUMPTION, False

    return _ScanJobResultEnum.SESSION_ID_RESUMPTION, True



from operator import attrgetter

from dataclasses import dataclass

from sslyze.connection_helpers.tls_connection import NoCiphersAvailableBugInSSlyze
from sslyze.plugins.openssl_cipher_suites._cli_connector import _CipherSuitesCliConnector
from sslyze.plugins.openssl_cipher_suites._test_cipher_suite import (
    connect_with_cipher_suite,
    CipherSuiteRejectedByServer,
    CipherSuiteAcceptedByServer,
)
from sslyze.plugins.openssl_cipher_suites.cipher_suites import CipherSuitesRepository
from sslyze.plugins.plugin_base import (
    ScanCommandImplementation,
    ScanCommandResult,
    ScanJob,
    ScanCommandExtraArgument,
    ScanCommandWrongUsageError,
    ScanJobResult,
)
from typing import ClassVar, Optional
from typing import List

from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum


@dataclass(frozen=True)
class CipherSuitesScanResult(ScanCommandResult):
    """The result of testing a server for cipher suites with a specific version of SSL/TLS.

    Attributes:
        tls_version_used: The SSL/TLS version used to connect to the server.
        accepted_ciphers: The list of cipher suites supported supported by both SSLyze and the server.
        rejected_ciphers: The list of cipher suites supported by SSLyze that were rejected by the server.
    """

    tls_version_used: TlsVersionEnum
    accepted_cipher_suites: List[CipherSuiteAcceptedByServer]
    rejected_cipher_suites: List[CipherSuiteRejectedByServer]

    @property
    def is_tls_version_supported(self) -> bool:
        """Is the SSL/TLS version used to connect the server supported by it?
        """
        return True if self.accepted_cipher_suites else False


class _Sslv20CliConnector(_CipherSuitesCliConnector):

    _cli_option = "sslv2"
    _cli_description = "Test a server for SSL 2.0 support."
    _title_in_output = "SSL 2.0 Cipher Suites"


class _Sslv30CliConnector(_CipherSuitesCliConnector):

    _cli_option = "sslv3"
    _cli_description = "Test a server for SSL 3.0 support."
    _title_in_output = "SSL 3.0 Cipher Suites"


class _Tlsv10CliConnector(_CipherSuitesCliConnector):

    _cli_option = "tlsv1"
    _cli_description = "Test a server for TLS 1.0 support."
    _title_in_output = "TLS 1.0 Cipher Suites"


class _Tlsv11CliConnector(_CipherSuitesCliConnector):

    _cli_option = "tlsv1_1"
    _cli_description = "Test a server for TLS 1.1 support."
    _title_in_output = "TLS 1.1 Cipher Suites"


class _Tlsv12CliConnector(_CipherSuitesCliConnector):

    _cli_option = "tlsv1_2"
    _cli_description = "Test a server for TLS 1.2 support."
    _title_in_output = "TLS 1.2 Cipher Suites"


class _Tlsv13CliConnector(_CipherSuitesCliConnector):

    _cli_option = "tlsv1_3"
    _cli_description = "Test a server for TLS 1.3 support."
    _title_in_output = "TLS 1.3 Cipher Suites"


class _CipherSuitesScanImplementation(ScanCommandImplementation[CipherSuitesScanResult, None]):

    # The SSL version corresponding to the scan command
    _tls_version: ClassVar[TlsVersionEnum]

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        # Run one job per cipher suite to test for
        all_cipher_suites_to_test = CipherSuitesRepository.get_all_cipher_suites(cls._tls_version)
        scan_jobs = [
            ScanJob(
                function_to_call=connect_with_cipher_suite,
                function_arguments=[server_info, cls._tls_version, cipher_suite],
            )
            for cipher_suite in all_cipher_suites_to_test
        ]
        return scan_jobs

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> CipherSuitesScanResult:
        expected_scan_jobs_count = len(CipherSuitesRepository.get_all_cipher_suites(cls._tls_version))
        if len(scan_job_results) != expected_scan_jobs_count:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        accepted_cipher_suites = []
        rejected_cipher_suites = []
        for completed_job in scan_job_results:
            try:
                cipher_suite_result = completed_job.get_result()
            except NoCiphersAvailableBugInSSlyze:
                # Happens when we passed a cipher suite and a TLS version that are not supported together by OpenSSL
                # Swallowing this exception makes it easier as we can just always use the ALL:COMPLEMENTOFALL OpenSSL
                # cipher string instead of having to figure out exactly which cipher suites are supported by which
                # versions
                continue
            if isinstance(cipher_suite_result, CipherSuiteAcceptedByServer):
                accepted_cipher_suites.append(cipher_suite_result)
            elif isinstance(cipher_suite_result, CipherSuiteRejectedByServer):
                rejected_cipher_suites.append(cipher_suite_result)
            else:
                raise ValueError("Should never happen")

        # Sort all the lists
        accepted_cipher_suites.sort(key=attrgetter("cipher_suite.name"), reverse=True)
        rejected_cipher_suites.sort(key=attrgetter("cipher_suite.name"), reverse=True)

        # Generate the results
        return CipherSuitesScanResult(
            tls_version_used=cls._tls_version,
            accepted_cipher_suites=accepted_cipher_suites,
            rejected_cipher_suites=rejected_cipher_suites,
        )


class Sslv20ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for SSL 2.0 support.
    """

    cli_connector_cls = _Sslv20CliConnector
    _tls_version = TlsVersionEnum.SSL_2_0


class Sslv30ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for SSL 3.0 support.
    """

    cli_connector_cls = _Sslv30CliConnector
    _tls_version = TlsVersionEnum.SSL_3_0


class Tlsv10ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for TLS 1.0 support.
    """

    cli_connector_cls = _Tlsv10CliConnector
    _tls_version = TlsVersionEnum.TLS_1_0


class Tlsv11ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for TLS 1.1 support.
    """

    cli_connector_cls = _Tlsv11CliConnector
    _tls_version = TlsVersionEnum.TLS_1_1


class Tlsv12ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for TLS 1.2 support.
    """

    cli_connector_cls = _Tlsv12CliConnector
    _tls_version = TlsVersionEnum.TLS_1_2


class Tlsv13ScanImplementation(_CipherSuitesScanImplementation):
    """Test a server for TLS 1.3 support.
    """

    cli_connector_cls = _Tlsv13CliConnector
    _tls_version = TlsVersionEnum.TLS_1_3



from typing import List, TYPE_CHECKING, ClassVar

from nassl.ephemeral_key_info import EcDhEphemeralKeyInfo, DhEphemeralKeyInfo

from sslyze.plugins.openssl_cipher_suites._test_cipher_suite import CipherSuiteAcceptedByServer
from sslyze.plugins.plugin_base import ScanCommandCliConnector
from sslyze.server_connectivity import TlsVersionEnum

if TYPE_CHECKING:
    from sslyze.plugins.openssl_cipher_suites.implementation import CipherSuitesScanResult


class _CipherSuitesCliConnector(ScanCommandCliConnector["CipherSuitesScanResult", None]):

    _title_in_output: ClassVar[str]

    @classmethod
    def result_to_console_output(cls, result: "CipherSuitesScanResult") -> List[str]:
        result_as_txt = [cls._format_title(cls._title_in_output)]

        cipher_suites_count = len(result.accepted_cipher_suites) + len(result.rejected_cipher_suites)
        if not result.accepted_cipher_suites:
            result_as_txt.append(
                cls._format_subtitle(
                    f"Attempted to connect using {cipher_suites_count} cipher suites;"
                    f" the server rejected all cipher suites."
                )
            )
            return result_as_txt

        # Display all cipher suites that were accepted
        result_as_txt.append(cls._format_subtitle(f"Attempted to connect using {cipher_suites_count} cipher suites."))
        result_as_txt.append("")
        result_as_txt.append(
            cls._format_subtitle(
                f"The server accepted the following {len(result.accepted_cipher_suites)} cipher suites:"
            )
        )
        for accepted_cipher in result.accepted_cipher_suites:
            result_as_txt.append(_format_accepted_cipher_suite(accepted_cipher))
        result_as_txt.append("")

        # Display some general comments about the cipher suite configuration
        # These comments only apply to TLS versions below 1.3 because TLS 1.3 removed "bad" cipher suites
        if result.tls_version_used.value < TlsVersionEnum.TLS_1_3.value:
            result_as_txt.append(
                cls._format_subtitle("The group of cipher suites supported by the server has the following properties:")
            )

            # Forward secrecy
            supports_forward_secrecy = False
            for accepted_cipher in result.accepted_cipher_suites:
                if "_DHE_" in accepted_cipher.cipher_suite.name or "_ECDHE_" in accepted_cipher.cipher_suite.name:
                    supports_forward_secrecy = True
                    break

            result_as_txt.append(
                cls._format_field(
                    "Forward Secrecy", "OK - Supported" if supports_forward_secrecy else "INSECURE - Not Supported"
                )
            )

            # Insecure RC4 cipher suites
            supports_rc4 = False
            for accepted_cipher in result.accepted_cipher_suites:
                if "_RC4_" in accepted_cipher.cipher_suite.name:
                    supports_rc4 = True
                    break
            result_as_txt.append(
                cls._format_field(
                    "Legacy RC4 Algorithm", "INSECURE - Supported" if supports_rc4 else "OK - Not Supported"
                )
            )
            result_as_txt.append("")

        return result_as_txt


def _format_accepted_cipher_suite(accepted_cipher: CipherSuiteAcceptedByServer) -> str:
    eph_key = accepted_cipher.ephemeral_key
    if isinstance(eph_key, EcDhEphemeralKeyInfo):
        dh_info = f"ECDH: {eph_key.curve_name} ({eph_key.size} bits)"
    elif isinstance(eph_key, DhEphemeralKeyInfo):
        dh_info = f"DH ({eph_key.size} bits)"
    else:
        dh_info = ""

    return f"        {accepted_cipher.cipher_suite.name:<50}{accepted_cipher.cipher_suite.key_size:<10}{dh_info:<15}"



from dataclasses import dataclass

import pydantic
from nassl.legacy_ssl_client import LegacySslClient
from nassl.ssl_client import ClientCertificateRequested

from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.plugins.plugin_base import (
    ScanCommandResult,
    ScanCommandImplementation,
    ScanJob,
    ScanCommandExtraArgument,
    ScanCommandWrongUsageError,
    ScanCommandCliConnector,
    ScanJobResult,
)
from typing import List, Optional

from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum
from sslyze.errors import ServerRejectedTlsHandshake


@dataclass(frozen=True)
class CompressionScanResult(ScanCommandResult):
    """The result of testing a server for TLS compression support.

    Attributes:
        supports_compression: True if TLS compression is supported by the server, thereby enabling the CRIME attack.
    """

    supports_compression: bool


# Identical fields in the JSON output
CompressionScanResultAsJson = pydantic.dataclasses.dataclass(CompressionScanResult, frozen=True)


class CompressionScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[CompressionScanResultAsJson]  # type: ignore


class _CompressionCliConnector(ScanCommandCliConnector[CompressionScanResult, None]):

    _cli_option = "compression"
    _cli_description = "Test a server for TLS compression support, which can be leveraged to perform a CRIME attack."

    @classmethod
    def result_to_console_output(cls, result: CompressionScanResult) -> List[str]:
        result_as_txt = [cls._format_title("Deflate Compression")]
        if result.supports_compression:
            result_as_txt.append(cls._format_field("", "VULNERABLE - Server supports Deflate compression"))
        else:
            result_as_txt.append(cls._format_field("", "OK - Compression disabled"))
        return result_as_txt


class CompressionImplementation(ScanCommandImplementation[CompressionScanResult, None]):
    """Test a server for TLS compression support, which can be leveraged to perform a CRIME attack.
    """

    cli_connector_cls = _CompressionCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        return [ScanJob(function_to_call=_test_compression_support, function_arguments=[server_info])]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> CompressionScanResult:
        if len(scan_job_results) != 1:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        return CompressionScanResult(supports_compression=scan_job_results[0].get_result())


def _test_compression_support(server_info: ServerConnectivityInfo) -> bool:
    # Try with TLS 1.2 even if the server supports TLS 1.3 or higher as there is no compression with TLS 1.3
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        tls_version_to_use = TlsVersionEnum.TLS_1_2
        downgraded_from_tls_1_3 = True
    else:
        tls_version_to_use = server_info.tls_probing_result.highest_tls_version_supported
        downgraded_from_tls_1_3 = False

    ssl_connection = server_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version_to_use,
        should_use_legacy_openssl=True,  # Only the legacy SSL client has methods to check for compression support
    )
    if not isinstance(ssl_connection.ssl_client, LegacySslClient):
        raise RuntimeError("Should never happen")

    # Make sure OpenSSL was built with support for compression to avoid false negatives
    if "zlib compression" not in ssl_connection.ssl_client.get_available_compression_methods():
        raise RuntimeError("OpenSSL was not built with support for zlib / compression. Did you build nassl yourself ?")

    compression_name: Optional[str]
    try:
        # Perform the TLS handshake
        ssl_connection.connect()
        compression_name = ssl_connection.ssl_client.get_current_compression_method()

    except ClientCertificateRequested:
        compression_name = ssl_connection.ssl_client.get_current_compression_method()

    # Should only happen when the server only supports TLS 1.3, which does not support compression
    except ServerRejectedTlsHandshake:
        if downgraded_from_tls_1_3:
            compression_name = None
        else:
            raise

    finally:
        ssl_connection.close()

    return True if compression_name else False



from dataclasses import dataclass, asdict
from operator import attrgetter
from typing import List, Optional

import pydantic
from nassl._nassl import OpenSSLError
from nassl.ephemeral_key_info import OpenSslEcNidEnum, EcDhEphemeralKeyInfo, _OPENSSL_NID_TO_SECG_ANSI_X9_62
from nassl.ssl_client import ClientCertificateRequested, SslClient

from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.server_connectivity import ServerConnectivityInfo
from sslyze.errors import ServerRejectedTlsHandshake, TlsHandshakeTimedOut
from sslyze.plugins.plugin_base import (
    ScanCommandResult,
    ScanCommandCliConnector,
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandWrongUsageError,
    ScanJobResult,
)
from sslyze.server_connectivity import enable_ecdh_cipher_suites


@dataclass(frozen=True)
class EllipticCurve:
    """A specific elliptic curve.

    Attributes:
        name: The ANSI X9.62 name if available, otherwise the SECG name.
        openssl_nid: The OpenSSL NID_XXX value valid for OpenSslEvpPkeyEnum.EC (obj_mac.h).
    """

    name: str
    openssl_nid: int


@dataclass(frozen=True)
class SupportedEllipticCurvesScanResult(ScanCommandResult):
    """The result of testing a server for supported elliptic curves.

    Attributes:
        supports_ecdh_key_exchange: True if the server supports at least one cipher suite with an ECDH key exchange.
        supported_curves: A list of `EllipticCurve` that were accepted by the server or `None` if the server does not
            support ECDH cipher suites.
        rejected_curves: A list of `EllipticCurve` that were rejected by the server or `None` if the server does not
            support ECDH cipher suites.
    """

    supports_ecdh_key_exchange: bool
    supported_curves: Optional[List[EllipticCurve]]
    rejected_curves: Optional[List[EllipticCurve]]

    def __post_init__(self) -> None:
        # Sort the curves by name
        if self.supported_curves:
            self.supported_curves.sort(key=attrgetter("name"))
        if self.rejected_curves:
            self.rejected_curves.sort(key=attrgetter("name"))


class _EllipticCurveAsJson(pydantic.BaseModel):
    name: str
    openssl_nid: int


_EllipticCurveAsJson.__doc__ = EllipticCurve.__doc__  # type: ignore


class SupportedEllipticCurvesScanResultAsJson(pydantic.BaseModel):
    supports_ecdh_key_exchange: bool
    supported_curves: Optional[List[_EllipticCurveAsJson]]
    rejected_curves: Optional[List[_EllipticCurveAsJson]]

    class Config:
        orm_mode = True

    @classmethod
    def from_orm(cls, result: SupportedEllipticCurvesScanResult) -> "SupportedEllipticCurvesScanResultAsJson":
        supported_curves: Optional[List[_EllipticCurveAsJson]] = None
        if result.supported_curves:
            supported_curves = [_EllipticCurveAsJson(**asdict(curve)) for curve in result.supported_curves]

        rejected_curves: Optional[List[_EllipticCurveAsJson]] = None
        if result.rejected_curves:
            rejected_curves = [_EllipticCurveAsJson(**asdict(curve)) for curve in result.rejected_curves]

        return cls(
            supports_ecdh_key_exchange=result.supports_ecdh_key_exchange,
            supported_curves=supported_curves,
            rejected_curves=rejected_curves,
        )


SupportedEllipticCurvesScanResultAsJson.__doc__ = SupportedEllipticCurvesScanResult.__doc__  # type: ignore


class SupportedEllipticCurvesScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[SupportedEllipticCurvesScanResultAsJson]


class _SupportedEllipticCurvesCliConnector(ScanCommandCliConnector[SupportedEllipticCurvesScanResult, None]):

    _cli_option = "elliptic_curves"
    _cli_description = "Test a server for supported elliptic curves."

    @classmethod
    def result_to_console_output(cls, result: SupportedEllipticCurvesScanResult) -> List[str]:
        result_as_txt = [cls._format_title("Elliptic Curve Key Exchange")]

        if not result.supports_ecdh_key_exchange:
            result_as_txt.append(
                cls._format_subtitle("The server does not support cipher suites with ECDH key exchanges.")
            )
        else:
            if result.supported_curves is None:
                raise RuntimeError("Should never happen")
            if result.rejected_curves is None:
                raise RuntimeError("Should never happen")

            supported_curves_names = [curve.name for curve in result.supported_curves]
            rejected_curves_names = [curve.name for curve in result.rejected_curves]
            result_as_txt.append(cls._format_field("Supported curves:", ", ".join(supported_curves_names)))
            result_as_txt.append(cls._format_field("Rejected curves:", ", ".join(rejected_curves_names)))
        return result_as_txt


class SupportedEllipticCurvesImplementation(ScanCommandImplementation[SupportedEllipticCurvesScanResult, None]):
    """Test a server for supported elliptic curves.
    """

    cli_connector_cls = _SupportedEllipticCurvesCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        if not server_info.tls_probing_result.supports_ecdh_key_exchange:
            # Nothing to test: the server doesn't support EC key exchange
            return [ScanJob(function_to_call=_raise_elliptic_curve_not_supported, function_arguments=[])]

        # List of curves are in https://tools.ietf.org/html/rfc4492#section-5.1.1 and
        # https://tools.ietf.org/html/rfc8446#section-4.2.7
        return [
            ScanJob(function_to_call=_test_curve, function_arguments=[server_info, curve_nid])
            for curve_nid in OpenSslEcNidEnum
        ]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> SupportedEllipticCurvesScanResult:
        if len(scan_job_results) < 1:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        if len(scan_job_results) == 1:
            try:
                scan_job_results[0].get_result()
                raise RuntimeError("Should never happen")
            except _EllipticCurveNotSupported:
                return SupportedEllipticCurvesScanResult(
                    supports_ecdh_key_exchange=False, supported_curves=None, rejected_curves=None,
                )
        else:
            all_ecdh_results = [scan_job.get_result() for scan_job in scan_job_results]
            return SupportedEllipticCurvesScanResult(
                supports_ecdh_key_exchange=True,
                supported_curves=[
                    ec_result.curve for ec_result in all_ecdh_results if ec_result.was_accepted_by_server
                ],
                rejected_curves=[
                    ec_result.curve for ec_result in all_ecdh_results if not ec_result.was_accepted_by_server
                ],
            )


class _EllipticCurveNotSupported(Exception):
    pass


def _raise_elliptic_curve_not_supported() -> None:
    raise _EllipticCurveNotSupported()


@dataclass(frozen=True)
class _EllipticCurveResult:
    curve: EllipticCurve
    was_accepted_by_server: bool


def _test_curve(server_info: ServerConnectivityInfo, curve_nid: OpenSslEcNidEnum) -> _EllipticCurveResult:
    if not server_info.tls_probing_result.supports_ecdh_key_exchange:
        raise RuntimeError("Should never happen")

    tls_version = server_info.tls_probing_result.highest_tls_version_supported
    ssl_connection = server_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version, should_use_legacy_openssl=False
    )
    if not isinstance(ssl_connection.ssl_client, SslClient):
        raise RuntimeError(
            "Should never happen: specified should_use_legacy_openssl=False but didn't get the modern SSL client"
        )

    # Set curve to test whether it is supported by the server
    enable_ecdh_cipher_suites(tls_version, ssl_connection.ssl_client)
    ssl_connection.ssl_client.set_groups([curve_nid])

    try:
        ssl_connection.connect()
        negotiated_ephemeral_key = ssl_connection.ssl_client.get_ephemeral_key()

    # Error handling here mis similar to test_cipher_suite.py
    except ClientCertificateRequested:
        negotiated_ephemeral_key = ssl_connection.ssl_client.get_ephemeral_key()

    except (TlsHandshakeTimedOut, ServerRejectedTlsHandshake):
        negotiated_ephemeral_key = None

    except OpenSSLError as e:
        # The following errors can be triggered by some servers when they don't support the specific curve enabled
        # in the client
        if "ossl_statem_client_read_transition:unexpected message" in e.args[0]:
            # Related to https://github.com/nabla-c0d3/sslyze/issues/466
            negotiated_ephemeral_key = None
        elif "tls_process_ske_ecdhe:wrong curve" in e.args[0]:
            # https://github.com/nabla-c0d3/sslyze/issues/490
            negotiated_ephemeral_key = None
        elif "sslv3 alert unexpected message" in e.args[0]:
            # https://github.com/nabla-c0d3/sslyze/issues/490
            negotiated_ephemeral_key = None
        else:
            raise

    finally:
        ssl_connection.close()

    # If no error occurred check if the curve was really used
    curve_name = _OPENSSL_NID_TO_SECG_ANSI_X9_62[curve_nid]  # TODO(AD): Make this public in nassl
    if negotiated_ephemeral_key:
        if isinstance(negotiated_ephemeral_key, EcDhEphemeralKeyInfo):
            if negotiated_ephemeral_key.curve != curve_nid:
                raise RuntimeError("Should never happen")

            return _EllipticCurveResult(
                curve=EllipticCurve(name=curve_name, openssl_nid=curve_nid.value), was_accepted_by_server=True,
            )

    return _EllipticCurveResult(
        curve=EllipticCurve(name=curve_name, openssl_nid=curve_nid.value), was_accepted_by_server=False,
    )



from hashlib import sha256
from typing import List, cast

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.x509 import ExtensionOID, DNSName, ExtensionNotFound, NameOID
from cryptography.x509.extensions import DuplicateExtension  # type: ignore


def extract_dns_subject_alternative_names(certificate: x509.Certificate) -> List[str]:
    """Retrieve all the DNS entries of the Subject Alternative Name extension.
    """
    subj_alt_names: List[str] = []
    try:
        san_ext = certificate.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
        san_ext_value = cast(x509.SubjectAlternativeName, san_ext.value)
        subj_alt_names = san_ext_value.get_values_for_type(DNSName)
    except ExtensionNotFound:
        pass
    except DuplicateExtension:
        # Fix for https://github.com/nabla-c0d3/sslyze/issues/420
        # Not sure how browsers behave in this case but having a duplicate extension makes the certificate invalid
        # so we just return no SANs (likely to make hostname validation fail, which is fine)
        pass

    return subj_alt_names


def get_common_names(name_field: x509.Name) -> List[str]:
    return [cn.value for cn in name_field.get_attributes_for_oid(NameOID.COMMON_NAME)]


def get_public_key_sha256(certificate: x509.Certificate) -> bytes:
    pub_bytes = certificate.public_key().public_bytes(encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo)
    digest = sha256(pub_bytes).digest()
    return digest



from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sslyze.server_setting import ServerNetworkLocation, ServerNetworkConfiguration


class InvalidServerNetworkConfigurationError(Exception):
    """Raised when trying to create a ServerNetworkConfiguration with invalid settings.
    """


class ServerHostnameCouldNotBeResolved(Exception):
    """Raised when trying to create a ServerNetworkLocationViaDirectConnection with a hostname whose DNS lookup failed.
    """


@dataclass(frozen=True)
class ConnectionToServerFailed(Exception):
    """Parent class for all exceptions raised when a connecting to a server failed.
    """

    server_location: "ServerNetworkLocation"
    network_configuration: "ServerNetworkConfiguration"
    error_message: str

    def __str__(self) -> str:
        return f'{self.server_location.display_string} -> "{self.error_message}".'


@dataclass(frozen=True)
class ConnectionToServerTimedOut(ConnectionToServerFailed):
    pass


@dataclass(frozen=True)
class ServerRejectedConnection(ConnectionToServerFailed):
    pass


@dataclass(frozen=True)
class ConnectionToHttpProxyFailed(ConnectionToServerFailed):
    pass


@dataclass(frozen=True)
class ConnectionToHttpProxyTimedOut(ConnectionToHttpProxyFailed):
    pass


@dataclass(frozen=True)
class HttpProxyRejectedConnection(ConnectionToHttpProxyFailed):
    pass


@dataclass(frozen=True)
class ServerRejectedOpportunisticTlsNegotiation(ConnectionToServerFailed):
    pass


@dataclass(frozen=True)
class TlsHandshakeFailed(ABC, ConnectionToServerFailed):
    pass


@dataclass(frozen=True)
class ServerRejectedTlsHandshake(TlsHandshakeFailed):
    pass


@dataclass(frozen=True)
class ServerTlsConfigurationNotSupported(TlsHandshakeFailed):
    pass


@dataclass(frozen=True)
class TlsHandshakeTimedOut(TlsHandshakeFailed):
    """Raised when the initial socket connection to the server succeeded, but the TLS handshake then timed out.

    This means that the server is definitely reachable/online, but its TLS stack is buggy or it does not support the TLS
    versions SSLyze enabled in the handshake.

    See https://github.com/nabla-c0d3/sslyze/issues/445 for more details.
    """

    pass



from dataclasses import dataclass
from pathlib import Path

from cryptography.x509.base import Certificate
from cryptography.x509.extensions import ExtensionNotFound, CertificatePolicies
from cryptography.x509.oid import ObjectIdentifier
from cryptography.x509.oid import ExtensionOID
from typing import List, cast
from typing import Optional


@dataclass(frozen=True)
class TrustStore:
    """A set of root certificates to be used for certificate validation.

    Attributes:
        path: The path on the local system to the PEM-formatted file containing the root certificates.
        name: The human-readable name of the trust store (such as "Mozilla").
        version: The human-readable version or date of the trust store (such as "09/2016").
    """

    path: Path
    name: str
    version: str
    ev_oids: Optional[List[ObjectIdentifier]] = None

    def is_certificate_extended_validation(self, certificate: Certificate) -> bool:
        """Is the supplied server certificate EV?
        """
        if not self.ev_oids:
            raise ValueError("No EV OIDs supplied for {} store - cannot detect EV certificates".format(self.name))

        try:
            cert_policies_ext = certificate.extensions.get_extension_for_oid(ExtensionOID.CERTIFICATE_POLICIES)
        except ExtensionNotFound:
            return False

        cert_policies_value = cast(CertificatePolicies, cert_policies_ext.value)
        for policy in cert_policies_value:
            if policy.policy_identifier in self.ev_oids:
                return True
        return False



from dataclasses import dataclass
from typing import Optional, List, Dict

import pydantic

from sslyze.json.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.plugins.plugin_base import (
    ScanCommandResult,
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandWrongUsageError,
    ScanCommandCliConnector,
    ScanJobResult,
)

from sslyze.plugins.robot._robot_tester import (
    RobotScanResultEnum,
    test_robot,
    ServerDoesNotSupportRsa,
    RobotPmsPaddingPayloadEnum,
    RobotServerResponsesAnalyzer,
)
from sslyze.server_connectivity import ServerConnectivityInfo


@dataclass(frozen=True)
class RobotScanResult(ScanCommandResult):
    """The result of testing a server for the ROBOT vulnerability.

    Attributes:
        result: An Enum providing the result of the ROBOT scan.
    """

    robot_result: RobotScanResultEnum


# Identical fields in the JSON output
RobotScanResultAsJson = pydantic.dataclasses.dataclass(RobotScanResult, frozen=True)


class RobotScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[RobotScanResultAsJson]  # type: ignore


class _RobotCliConnector(ScanCommandCliConnector[RobotScanResult, None]):

    _cli_option = "robot"
    _cli_description = "Test a server for the ROBOT vulnerability."

    @classmethod
    def result_to_console_output(cls, result: RobotScanResult) -> List[str]:
        result_as_txt = [cls._format_title("ROBOT Attack")]

        if result.robot_result == RobotScanResultEnum.VULNERABLE_STRONG_ORACLE:
            robot_txt = "VULNERABLE - Strong oracle, a real attack is possible."
        elif result.robot_result == RobotScanResultEnum.VULNERABLE_WEAK_ORACLE:
            robot_txt = "VULNERABLE - Weak oracle, the attack would take too long."
        elif result.robot_result == RobotScanResultEnum.NOT_VULNERABLE_NO_ORACLE:
            robot_txt = "OK - Not vulnerable."
        elif result.robot_result == RobotScanResultEnum.NOT_VULNERABLE_RSA_NOT_SUPPORTED:
            robot_txt = "OK - Not vulnerable, RSA cipher suites not supported."
        elif result.robot_result == RobotScanResultEnum.UNKNOWN_INCONSISTENT_RESULTS:
            robot_txt = "UNKNOWN - Received inconsistent results."
        else:
            raise ValueError("Should never happen")
        result_as_txt.append(cls._format_field("", robot_txt))
        return result_as_txt


class RobotImplementation(ScanCommandImplementation[RobotScanResult, None]):
    """Test a server for the ROBOT vulnerability.
    """

    cli_connector_cls = _RobotCliConnector

    _TEST_ATTEMPTS_NB = 3

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        # Run the test three times to ensure the results are consistent
        return [
            ScanJob(function_to_call=test_robot, function_arguments=[server_info]) for _ in range(cls._TEST_ATTEMPTS_NB)
        ]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> RobotScanResult:
        if len(scan_job_results) != cls._TEST_ATTEMPTS_NB:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        combined_server_responses: Dict[RobotPmsPaddingPayloadEnum, List[str]] = {
            payload_enum: [] for payload_enum in RobotPmsPaddingPayloadEnum
        }
        for future in scan_job_results:
            try:
                server_responses_per_robot_payloads = future.get_result()
                for payload_enum, server_response in server_responses_per_robot_payloads.items():
                    combined_server_responses[payload_enum].append(server_response)
            except ServerDoesNotSupportRsa:
                return RobotScanResult(robot_result=RobotScanResultEnum.NOT_VULNERABLE_RSA_NOT_SUPPORTED)

        result = RobotServerResponsesAnalyzer(combined_server_responses, cls._TEST_ATTEMPTS_NB).compute_result_enum()
        return RobotScanResult(result)



from dataclasses import fields
from datetime import datetime
from pathlib import Path
from typing import TextIO, Optional

from sslyze import __version__, ServerScanRequest, ScanCommandAttemptStatusEnum, ScanCommandErrorReasonEnum
from sslyze.cli.command_line_parser import ParsedCommandLine

from sslyze.errors import ConnectionToServerFailed
from sslyze.plugins.plugin_base import ScanCommandWrongUsageError
from sslyze.plugins.scan_commands import ScanCommandsRepository, ScanCommand
from sslyze import ServerScanResult
from sslyze.scanner.models import ServerScanStatusEnum
from sslyze.scanner.scan_command_attempt import ScanCommandAttempt
from sslyze.scanner.scanner_observer import ScannerObserver
from sslyze.server_connectivity import ClientAuthRequirementEnum, ServerTlsProbingResult
from sslyze.server_setting import (
    ServerNetworkLocation,
    ConnectionTypeEnum,
)


class ObserverToGenerateConsoleOutput(ScannerObserver):
    def __init__(self, file_to: TextIO, json_path_out: Optional[Path] = None) -> None:
        self._file_to = file_to
        self._date_scans_started = datetime.utcnow()

        # Used to print the path where the JSON output was written
        self._json_path_out = json_path_out

    @classmethod
    def _format_title(cls, title: str) -> str:
        return f" {title.upper()}\n {'-' * len(title)}\n"

    def command_line_parsed(self, parsed_command_line: ParsedCommandLine) -> None:
        self._file_to.write("\n")
        self._file_to.write(self._format_title("Checking connectivity to server(s)"))
        self._file_to.write("\n")

        for bad_server_str in parsed_command_line.invalid_servers:
            self._file_to.write(
                f"   {bad_server_str.server_string:<35} => ERROR: {bad_server_str.error_message};"
                f" discarding scan.\n"
            )

    def server_connectivity_test_error(
        self, server_scan_request: ServerScanRequest, connectivity_error: ConnectionToServerFailed
    ) -> None:
        self._file_to.write(
            f"   {connectivity_error.server_location.display_string:<25}"
            f" => ERROR: {connectivity_error.error_message}; discarding scan.\n"
        )

    def server_connectivity_test_completed(
        self, server_scan_request: ServerScanRequest, connectivity_result: ServerTlsProbingResult
    ) -> None:
        client_auth_msg = ""
        client_auth_requirement = connectivity_result.client_auth_requirement
        if client_auth_requirement == ClientAuthRequirementEnum.REQUIRED:
            client_auth_msg = "  WARNING: Server REQUIRED client authentication, specific plugins will fail."
        elif client_auth_requirement == ClientAuthRequirementEnum.OPTIONAL:
            client_auth_msg = "  WARNING: Server requested optional client authentication"

        server_location = server_scan_request.server_location
        network_route = _server_location_to_network_route(server_location)
        self._file_to.write(f"   {server_location.display_string:<25} => {network_route} {client_auth_msg}\n")

    def server_scan_completed(self, server_scan_result: ServerScanResult) -> None:
        if server_scan_result.scan_status != ServerScanStatusEnum.COMPLETED:
            # Nothing to print here if the scan was not completed
            return

        # Generate the console output for each scan command
        scan_command_results_str = ""
        for result_field in fields(server_scan_result.scan_result):
            scan_command = ScanCommand(result_field.name)
            scan_command_attempt = getattr(server_scan_result.scan_result, scan_command, None)

            if scan_command_attempt.status == ScanCommandAttemptStatusEnum.COMPLETED:
                scan_command_results_str += "\n"
                cli_connector_cls = ScanCommandsRepository.get_implementation_cls(scan_command).cli_connector_cls
                for line in cli_connector_cls.result_to_console_output(scan_command_attempt.result):
                    scan_command_results_str += line + "\n"

            elif scan_command_attempt.status == ScanCommandAttemptStatusEnum.ERROR:
                scan_command_results_str += scan_command_error_as_console_output(
                    server_scan_result.server_location, scan_command, scan_command_attempt
                )

            elif scan_command_attempt.status == ScanCommandAttemptStatusEnum.NOT_SCHEDULED:
                pass

            else:
                raise ValueError("Should never happen")

        # Also display the server that was scanned
        server_location = server_scan_result.server_location
        network_route = _server_location_to_network_route(server_location)
        scan_txt = f"Scan Results For {server_location.display_string} - {network_route}"
        self._file_to.write("\n" + self._format_title(scan_txt) + scan_command_results_str + "\n")

    def all_server_scans_completed(self) -> None:
        scans_duration = datetime.utcnow() - self._date_scans_started
        self._file_to.write(
            self._format_title(f"Scans Completed in {scans_duration.seconds}.{scans_duration.microseconds} s")
        )
        if self._json_path_out:
            self._file_to.write(f'\n       Wrote JSON output to "{self._json_path_out}".\n')


def _server_location_to_network_route(server_location: ServerNetworkLocation) -> str:
    if server_location.connection_type == ConnectionTypeEnum.VIA_HTTP_PROXY:
        # We do not know the server's IP address if going through a proxy
        assert server_location.http_proxy_settings
        network_route = "HTTP proxy at {}:{}".format(
            server_location.http_proxy_settings.hostname, server_location.http_proxy_settings.port
        )
    elif server_location.connection_type == ConnectionTypeEnum.DIRECT:
        assert server_location.ip_address
        network_route = server_location.ip_address
    else:
        raise ValueError("Should never happen")
    return network_route


def scan_command_error_as_console_output(
    server_location: ServerNetworkLocation, scan_command: ScanCommand, scan_command_attempt: ScanCommandAttempt
) -> str:
    if not scan_command_attempt.error_trace:
        raise ValueError("Should never happen")

    target_result_str = "\n"
    cli_connector_cls = ScanCommandsRepository.get_implementation_cls(scan_command).cli_connector_cls

    if scan_command_attempt.error_reason == ScanCommandErrorReasonEnum.CLIENT_CERTIFICATE_NEEDED:
        target_result_str += cli_connector_cls._format_title(
            f"Client certificated required for --{cli_connector_cls._cli_option}"
        )
        target_result_str += " use --cert and --key to provide one.\n"

    elif scan_command_attempt.error_reason == ScanCommandErrorReasonEnum.CONNECTIVITY_ISSUE:
        target_result_str += cli_connector_cls._format_title(
            f"Connection timed out for --{cli_connector_cls._cli_option}"
        )
        target_result_str += " try using --slow_connection to reduce the impact on the server.\n"

    elif scan_command_attempt.error_reason == ScanCommandErrorReasonEnum.WRONG_USAGE:
        target_result_str += cli_connector_cls._format_title(f"Wrong usage for --{cli_connector_cls._cli_option}")
        # Extract the last line which contains the reason
        last_line = None
        for line in scan_command_attempt.error_trace.format(chain=False):
            last_line = line
        if last_line:
            exception_cls_in_trace = f"{ScanCommandWrongUsageError.__name__}:"
            if exception_cls_in_trace in last_line:
                details_text = last_line.split(exception_cls_in_trace)[1].strip()
                target_result_str += f"       {details_text}"
            else:
                target_result_str += f"       {last_line}"

    elif scan_command_attempt.error_reason == ScanCommandErrorReasonEnum.BUG_IN_SSLYZE:
        target_result_str += cli_connector_cls._format_title(f"Error when running --{cli_connector_cls._cli_option}")
        target_result_str += "\n"
        target_result_str += (
            "       You can open an issue at https://github.com/nabla-c0d3/sslyze/issues"
            " with the following information:\n\n"
        )
        target_result_str += f"       * SSLyze version: {__version__.__version__}\n"
        target_result_str += f"       * Server: {server_location.display_string}"
        target_result_str += f" - {_server_location_to_network_route(server_location)}\n"
        target_result_str += f"       * Scan command: {scan_command}\n\n"
        for line in scan_command_attempt.error_trace.format(chain=False):
            target_result_str += f"       {line}"
    else:
        raise ValueError("Should never happen")

    return target_result_str



import socket
from pathlib import Path
from typing import Optional, TYPE_CHECKING

from nassl.legacy_ssl_client import LegacySslClient

from sslyze.server_setting import (
    ServerNetworkLocation,
    ServerNetworkConfiguration,
    ConnectionTypeEnum,
)
from sslyze.errors import (
    ConnectionToServerTimedOut,
    ServerRejectedConnection,
    ConnectionToServerFailed,
    ConnectionToHttpProxyTimedOut,
    HttpProxyRejectedConnection,
    ConnectionToHttpProxyFailed,
    ServerRejectedOpportunisticTlsNegotiation,
    ServerRejectedTlsHandshake,
    ServerTlsConfigurationNotSupported,
    TlsHandshakeTimedOut,
)
from sslyze.connection_helpers.http_response_parser import HttpResponseParser

import time

from nassl import _nassl
from nassl.ssl_client import SslClient, OpenSslVersionEnum, BaseSslClient, OpenSslVerifyEnum
from nassl.ssl_client import ClientCertificateRequested

from sslyze.connection_helpers.opportunistic_tls_helpers import get_opportunistic_tls_helper, OpportunisticTlsError

if TYPE_CHECKING:
    from sslyze.server_connectivity import TlsVersionEnum


def _open_socket_for_direct_connection(server_location: ServerNetworkLocation, network_timeout: int) -> socket.socket:
    assert server_location.ip_address
    # Exceptions get caught/processed by the caller
    return socket.create_connection((server_location.ip_address, server_location.port), timeout=network_timeout)


class _ConnectionToHttpProxyTimedOut(Exception):
    pass


class _HttpProxyRejectedConnection(Exception):
    pass


class _ConnectionToHttpProxyFailed(Exception):
    pass


def _open_socket_for_connection_via_http_proxy(
    server_location: ServerNetworkLocation, network_timeout: int
) -> socket.socket:
    assert server_location.http_proxy_settings
    try:
        sock = socket.create_connection(
            (server_location.http_proxy_settings.hostname, server_location.http_proxy_settings.port),
            timeout=network_timeout,
        )

        # Send a CONNECT request with the host we want to tunnel to
        proxy_authorization_header = server_location.http_proxy_settings.proxy_authorization_header
        if proxy_authorization_header is None:
            sock.send(f"CONNECT {server_location.hostname}:{server_location.port} HTTP/1.1\r\n\r\n".encode("utf-8"))
        else:
            sock.send(
                (
                    f"CONNECT {server_location.hostname}:{server_location.port} HTTP/1.1\r\n"
                    f"Proxy-Authorization: Basic {proxy_authorization_header}\r\n\r\n"
                ).encode("utf-8")
            )
        http_response = HttpResponseParser.parse_from_socket(sock)
    except socket.timeout:
        raise _ConnectionToHttpProxyTimedOut()
    except ConnectionError:
        raise _HttpProxyRejectedConnection("The HTTP proxy rejected the connection")
    except socket.error:
        raise _ConnectionToHttpProxyFailed()

    # Check if the proxy was able to connect to the host
    if http_response.status != 200:
        raise _HttpProxyRejectedConnection("The HTTP proxy rejected the CONNECT request")

    return sock


def _open_socket(server_location: ServerNetworkLocation, network_timeout: int) -> socket.socket:
    if server_location.connection_type == ConnectionTypeEnum.VIA_HTTP_PROXY:
        return _open_socket_for_connection_via_http_proxy(server_location, network_timeout)
    elif server_location.connection_type == ConnectionTypeEnum.DIRECT:
        return _open_socket_for_direct_connection(server_location, network_timeout)
    else:
        raise ValueError()


_HANDSHAKE_REJECTED_TLS_ERRORS = {
    "excessive message size": "TLS error: excessive message size",
    "bad mac decode": "TLS error: bad mac decode",
    "wrong version number": "TLS error: wrong version number",
    "no cipher match": "TLS error: no cipher match",
    "bad decompression": "TLS error: bad decompression",
    "peer error no cipher": "TLS error: peer error no cipher",
    "no cipher list": "TLS error: no ciphers list",
    "insufficient security": "TLS error: insufficient security",
    "block type is not 01": "TLS error: block type is not 01",  # Actually an RSA error
    "wrong ssl version": "TLS error: wrong SSL version",
    "sslv3 alert handshake failure": "TLS alert: handshake failure",
    "tlsv1 alert protocol version": "TLS alert: protocol version ",
    "tlsv1 alert decrypt error": "TLS alert: Decrypt error",
    "tlsv1 alert decode error": "TLS alert: Decode error",
    "Connection was shut down by peer": "Server closed the connection during the TLS handshake",
    "alert bad record mac": "TLS alert: bad record mac",
    "tlsv1 alert internal error": "TLS alert: Internal error",
    "illegal padding": "TLS alert: Illegal padding",
    # Error returned by OpenSSL when the server didn't return a certificate that can work with the cipher suites
    # enabled in the client; for example client only supports EC cipher suites but server returned an RSA certificate
    "wrong certificate type": "Server returned wrong certificate type",
}


class NoCiphersAvailableBugInSSlyze(Exception):
    """Should never happen.
    """


class SslConnection:
    """SSL connection that handles error processing, including retries when receiving timeouts.

    This it the base class to use to connect to a server in order to scan it.
    """

    def __init__(
        self,
        server_location: ServerNetworkLocation,
        network_configuration: ServerNetworkConfiguration,
        tls_version: "TlsVersionEnum",
        should_ignore_client_auth: bool,
        should_use_legacy_openssl: Optional[bool] = None,
        ca_certificates_path: Optional[Path] = None,
        should_enable_server_name_indication: bool = True,
    ) -> None:
        self._server_location = server_location
        self._network_configuration = network_configuration

        # Create the SSL client
        nassl_tls_version = OpenSslVersionEnum(tls_version.value)
        self.ssl_client: BaseSslClient
        # For older versions of TLS/SSL, we have to use a legacy OpenSSL
        final_should_use_legacy_openssl: bool
        if should_use_legacy_openssl is None:
            # For older versions of TLS/SSL, we have to use a legacy OpenSSL
            final_should_use_legacy_openssl = (
                False if nassl_tls_version in [OpenSslVersionEnum.TLSV1_2, OpenSslVersionEnum.TLSV1_3] else True
            )
        else:
            final_should_use_legacy_openssl = should_use_legacy_openssl

        if nassl_tls_version == OpenSslVersionEnum.TLSV1_3 and final_should_use_legacy_openssl:
            raise ValueError("Cannot use legacy OpenSSL with TLS 1.3")
        elif (
            nassl_tls_version in [OpenSslVersionEnum.SSLV2, OpenSslVersionEnum.SSLV3]
            and not final_should_use_legacy_openssl
        ):
            raise ValueError("Cannot use modern OpenSSL with SSL 2.0 or 3.0")

        ssl_client_cls = LegacySslClient if final_should_use_legacy_openssl else SslClient

        if network_configuration.tls_client_auth_credentials:
            # A client certificate and private key were provided
            self.ssl_client = ssl_client_cls(
                ssl_version=nassl_tls_version,
                ssl_verify=OpenSslVerifyEnum.NONE,
                ssl_verify_locations=ca_certificates_path,
                client_certificate_chain=network_configuration.tls_client_auth_credentials.certificate_chain_path,
                client_key=network_configuration.tls_client_auth_credentials.key_path,
                client_key_type=network_configuration.tls_client_auth_credentials.key_type,
                client_key_password=network_configuration.tls_client_auth_credentials.key_password,
                ignore_client_authentication_requests=False,
            )
        else:
            # No client cert and key
            self.ssl_client = ssl_client_cls(
                ssl_version=nassl_tls_version,
                ssl_verify=OpenSslVerifyEnum.NONE,
                ssl_verify_locations=ca_certificates_path,
                ignore_client_authentication_requests=should_ignore_client_auth,
            )

        # And a default cipher list to make the client hello smaller so we don't run into
        # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=665452
        if nassl_tls_version != OpenSslVersionEnum.TLSV1_3:
            self.ssl_client.set_cipher_list("HIGH:MEDIUM:-aNULL:-eNULL:-3DES:-SRP:-PSK:-CAMELLIA")

        # Add Server Name Indication
        if should_enable_server_name_indication and nassl_tls_version != OpenSslVersionEnum.SSLV2:
            # TODO(AD): Modify set_tlsext_host_name() to return an exception so we dont need to look at _ssl_version
            self.ssl_client.set_tlsext_host_name(network_configuration.tls_server_name_indication)

    def _do_pre_handshake(self) -> None:
        try:
            sock = _open_socket(self._server_location, self._network_configuration.network_timeout)
        # Re-raise any proxy error with additional context/info
        except _ConnectionToHttpProxyTimedOut:
            raise ConnectionToHttpProxyTimedOut(
                server_location=self._server_location,
                network_configuration=self._network_configuration,
                error_message="Connection to HTTP Proxy timed out",
            )
        except _HttpProxyRejectedConnection as e:
            raise HttpProxyRejectedConnection(
                server_location=self._server_location,
                network_configuration=self._network_configuration,
                error_message=e.args[0],
            )
        except _ConnectionToHttpProxyFailed:
            raise ConnectionToHttpProxyFailed(
                server_location=self._server_location,
                network_configuration=self._network_configuration,
                error_message="Connection to the HTTP proxy failed",
            )

        # Do the Opportunistic/StartTLS negotiation if needed
        if self._network_configuration.tls_opportunistic_encryption:
            opportunistic_tls_helper = get_opportunistic_tls_helper(
                self._network_configuration.tls_opportunistic_encryption, self._network_configuration.xmpp_to_hostname
            )
            try:
                opportunistic_tls_helper.prepare_socket_for_tls_handshake(sock)
            except OpportunisticTlsError as e:
                raise ServerRejectedOpportunisticTlsNegotiation(
                    server_location=self._server_location,
                    error_message=e.args[0],
                    network_configuration=self._network_configuration,
                )

        # Pass the connected socket to the SSL client
        self.ssl_client.set_underlying_socket(sock)

    def connect(self, should_retry_connection: bool = True) -> None:
        max_attempts_nb = self._network_configuration.network_max_retries if should_retry_connection else 1
        connection_attempts_nb = 0
        delay_for_next_attempt = 0

        # First try to connect to the server, and do retries if there are timeouts
        while True:
            # Sleep if it's a retry attempt
            time.sleep(delay_for_next_attempt)
            try:
                self._do_pre_handshake()
            except socket.timeout:
                # Attempt to retry connection if a network error occurred during connection or the handshake
                connection_attempts_nb += 1
                if connection_attempts_nb >= max_attempts_nb:
                    # Exhausted the number of retry attempts, give up
                    raise ConnectionToServerTimedOut(
                        server_location=self._server_location,
                        network_configuration=self._network_configuration,
                        error_message="Connection to the server timed out",
                    )
                elif connection_attempts_nb == 1:
                    # Start with a 1 second delay
                    delay_for_next_attempt = 1
                else:
                    # Exponential back off; cap maximum delay at 6 seconds
                    delay_for_next_attempt = min(6, 2 * delay_for_next_attempt)
            except ConnectionError:
                raise ServerRejectedConnection(
                    server_location=self._server_location,
                    network_configuration=self._network_configuration,
                    error_message="Server rejected the connection",
                )
            except OSError:
                # OSError is the parent class of all socket (ie. non-TLS) connection errors such as socket.timeout or
                # ConnectionError; hence this is the most generic error handler and should always be defined last
                raise ConnectionToServerFailed(
                    server_location=self._server_location,
                    network_configuration=self._network_configuration,
                    error_message="Connection to the server failed",
                )

            else:
                # No network error occurred
                break

        # After successfully connecting to the server, perform the TLS handshake
        try:
            self.ssl_client.do_handshake()

        except ClientCertificateRequested:
            # Server expected a client certificate and we didn't provide one
            raise
        except socket.timeout:
            # Network timeout, propagate the error
            raise TlsHandshakeTimedOut(
                server_location=self._server_location,
                network_configuration=self._network_configuration,
                error_message="Connection to server timed out during the TLS handshake",
            )
        except ConnectionError:
            raise ServerRejectedTlsHandshake(
                server_location=self._server_location,
                network_configuration=self._network_configuration,
                error_message="Server rejected the connection",
            )
        except OSError as e:
            # OSError is the parent of all (non-TLS) socket/connection errors so it should be last
            if "Nassl SSL handshake failed" in e.args[0]:
                # Special error returned by nassl
                raise ServerRejectedTlsHandshake(
                    server_location=self._server_location,
                    network_configuration=self._network_configuration,
                    error_message="Server interrupted the TLS handshake",
                )
            # Unknown connection error
            raise
        except _nassl.OpenSSLError as e:
            openssl_error_message = e.args[0]
            if "dh key too small" in openssl_error_message:
                # This is when SSLyze's OpenSSL rejects DH parameters (to protect against Logjam); this actually
                # means the server supports whatever cipher suite was used
                raise ServerTlsConfigurationNotSupported(
                    server_location=self._server_location,
                    network_configuration=self._network_configuration,
                    error_message="DH key too small",
                )

            if "no ciphers available" in openssl_error_message:
                # This one is returned by OpenSSL when a cipher set via set_cipher_list() is not actually supported
                # Should never happen (SSLyze bugs)
                raise NoCiphersAvailableBugInSSlyze(
                    f"Set a cipher that is not supported by nassl: {self.ssl_client.get_cipher_list()}"
                )

            for error_msg in _HANDSHAKE_REJECTED_TLS_ERRORS.keys():
                if error_msg in openssl_error_message:
                    raise ServerRejectedTlsHandshake(
                        server_location=self._server_location,
                        network_configuration=self._network_configuration,
                        error_message=_HANDSHAKE_REJECTED_TLS_ERRORS[error_msg],
                    )

            # Unknown SSL error if we get there
            raise

    def close(self) -> None:
        self.ssl_client.shutdown()



"""Main abstract plugin classes from which all the plugins should inherit.
"""

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor

from dataclasses import dataclass

from typing import (
    List,
    Callable,
    Any,
    Optional,
    TYPE_CHECKING,
    Tuple,
    ClassVar,
    Dict,
    Type,
    Union,
    TypeVar,
    Generic,
    Sequence,
)

if TYPE_CHECKING:
    from sslyze.server_connectivity import ServerConnectivityInfo


class ScanCommandResult(ABC):
    pass


class ScanCommandExtraArgument(ABC):
    pass


class ScanCommandWrongUsageError(Exception):
    """Raised when the configuration or arguments passed to complete a scan command are wrong.
    """

    pass


@dataclass(frozen=True)
class ScanJob:
    """One scan job should encapsulate some kind of server testing that uses at most one network connection.

    This allows sslyze to accurately limit how many concurrent connections it opens to a single server.
    """

    function_to_call: Callable
    function_arguments: Sequence[Any]


@dataclass(frozen=True)
class ScanJobResult:
    _return_value: Optional[Any]
    _exception: Optional[Exception]

    def get_result(self) -> Any:
        if self._exception:
            raise self._exception
        else:
            return self._return_value


_ScanCommandResultTypeVar = TypeVar("_ScanCommandResultTypeVar", bound=ScanCommandResult)
_ScanCommandExtraArgumentTypeVar = TypeVar("_ScanCommandExtraArgumentTypeVar", bound=Optional[ScanCommandExtraArgument])


class ScanCommandImplementation(Generic[_ScanCommandResultTypeVar, _ScanCommandExtraArgumentTypeVar]):
    """Describes everything needed to run a specific scan command.
    """

    # Contains all the logic for making the scan command available via the CLI
    cli_connector_cls: ClassVar[Type["ScanCommandCliConnector"]]

    @classmethod
    @abstractmethod
    def scan_jobs_for_scan_command(
        cls, server_info: "ServerConnectivityInfo", extra_arguments: Optional[_ScanCommandExtraArgumentTypeVar] = None
    ) -> List[ScanJob]:
        """Transform a scan command to run into smaller scan jobs to be run concurrently.

        To ensure reliability of the scans, each job should use at most one network connection to the server that is
        being scanned.
        """
        pass

    @classmethod
    @abstractmethod
    def result_for_completed_scan_jobs(
        cls, server_info: "ServerConnectivityInfo", scan_job_results: List[ScanJobResult]
    ) -> _ScanCommandResultTypeVar:
        """Transform the individual scan job results for a given scan command into a scan command result.
        """
        pass

    @classmethod
    def scan_server(
        cls, server_info: "ServerConnectivityInfo", extra_arguments: Optional[_ScanCommandExtraArgumentTypeVar] = None
    ) -> _ScanCommandResultTypeVar:
        """Utility method to run a scan command directly.

        This is useful for the test suite to run commands without using the Scanner class. It should NOT be used to
        actually run scans as this will be very slow (no multi-threading); use the Scanner class instead.
        """
        thread_pool = ThreadPoolExecutor(max_workers=5)

        all_jobs = cls.scan_jobs_for_scan_command(server_info, extra_arguments)
        all_job_results = []
        for job in all_jobs:
            future = thread_pool.submit(job.function_to_call, *job.function_arguments)
            try:
                job_result = ScanJobResult(_return_value=future.result(), _exception=None)
            except Exception as e:
                job_result = ScanJobResult(_return_value=None, _exception=e)
            all_job_results.append(job_result)

        result = cls.result_for_completed_scan_jobs(server_info, all_job_results)
        return result


@dataclass(frozen=True)
class OptParseCliOption:
    option: str
    help: str
    action: str = "store_true"


class ScanCommandCliConnector(Generic[_ScanCommandResultTypeVar, _ScanCommandExtraArgumentTypeVar]):
    """Contains all the logic for making a scan command available via the CLI.
    """

    _cli_option: ClassVar[str]
    _cli_description: ClassVar[str]

    @classmethod
    def get_cli_options(cls) -> List[OptParseCliOption]:
        """Return the CLI option(s) relevant to the scan command.
        """
        # Subclasses can add command line options for extra arguments here; by default scan commands don't have
        # extra arguments
        return [OptParseCliOption(option=cls._cli_option, help=cls._cli_description)]

    @classmethod
    def find_cli_options_in_command_line(
        cls, parsed_command_line: Dict[str, Union[None, bool, str]]
    ) -> Tuple[bool, Optional[_ScanCommandExtraArgumentTypeVar]]:
        """Check a parsed command line to see if the CLI option for the scan command was enabled.
        """
        try:
            option = parsed_command_line[cls._cli_option]
            is_scan_cmd_enabled = True if option else False
        except KeyError:
            is_scan_cmd_enabled = False

        extra_arguments = None
        return is_scan_cmd_enabled, extra_arguments

    @classmethod
    @abstractmethod
    def result_to_console_output(cls, result: _ScanCommandResultTypeVar) -> List[str]:
        """Transform the result of the scan command into lines of text to be printed by the CLI.
        """
        pass

    # Common formatting methods to have a consistent console output
    @staticmethod
    def _format_title(title: str) -> str:
        return " * {0}:".format(title)

    @staticmethod
    def _format_subtitle(subtitle: str) -> str:
        return "     {0}".format(subtitle)

    @staticmethod
    def _format_field(title: str, value: str = "") -> str:
        return "       {0:<35}{1}".format(title, value)



import sys
from datetime import datetime
from typing import Optional, TextIO

from sslyze.cli.console_output import ObserverToGenerateConsoleOutput
from sslyze.__version__ import __version__
from sslyze.cli.command_line_parser import CommandLineParsingError, CommandLineParser

from sslyze import (
    Scanner,
    ServerScanRequest,
    SslyzeOutputAsJson,
    ServerScanResultAsJson,
)
from sslyze.mozilla_tls_profile.mozilla_config_checker import (
    MozillaTlsConfigurationChecker,
    ServerNotCompliantWithMozillaTlsConfiguration,
    ServerScanResultIncomplete,
)


def main() -> None:
    # Parse the supplied command line
    date_scans_started = datetime.utcnow()
    sslyze_parser = CommandLineParser(__version__)
    try:
        parsed_command_line = sslyze_parser.parse_command_line()
    except CommandLineParsingError as e:
        print(e.get_error_msg())
        return

    # Setup the observer to print to the console, if needed
    scanner_observers = []
    if not parsed_command_line.should_disable_console_output:
        observer_for_console_output = ObserverToGenerateConsoleOutput(
            file_to=sys.stdout, json_path_out=parsed_command_line.json_path_out
        )
        observer_for_console_output.command_line_parsed(parsed_command_line=parsed_command_line)

        scanner_observers.append(observer_for_console_output)

    # Setup the scanner
    sslyze_scanner = Scanner(
        per_server_concurrent_connections_limit=parsed_command_line.per_server_concurrent_connections_limit,
        concurrent_server_scans_limit=parsed_command_line.concurrent_server_scans_limit,
        observers=scanner_observers,
    )

    # Queue the scans
    all_server_scan_requests = []
    for server_location, network_config in parsed_command_line.servers_to_scans:
        scan_request = ServerScanRequest(
            server_location=server_location,
            network_configuration=network_config,
            scan_commands=parsed_command_line.scan_commands,
            scan_commands_extra_arguments=parsed_command_line.scan_commands_extra_arguments,
        )
        all_server_scan_requests.append(scan_request)
    assert all_server_scan_requests

    sslyze_scanner.queue_scans(all_server_scan_requests)
    all_server_scan_results = []
    for result in sslyze_scanner.get_results():
        # Results are actually displayed by the observer; here we just store them
        all_server_scan_results.append(result)

    # Write results to a JSON file if needed
    json_file_out: Optional[TextIO] = None
    if parsed_command_line.should_print_json_to_console:
        json_file_out = sys.stdout
    elif parsed_command_line.json_path_out:
        json_file_out = parsed_command_line.json_path_out.open("wt", encoding="utf-8")

    if json_file_out:
        json_output = SslyzeOutputAsJson(
            server_scan_results=[ServerScanResultAsJson.from_orm(result) for result in all_server_scan_results],
            date_scans_started=date_scans_started,
            date_scans_completed=datetime.utcnow(),
        )
        json_output_as_str = json_output.json(sort_keys=True, indent=4, ensure_ascii=True)
        json_file_out.write(json_output_as_str)

    # Check the results against the Mozilla config if needed
    are_all_servers_compliant = True
    # TODO(AD): Expose format_title method
    title = ObserverToGenerateConsoleOutput._format_title("Compliance against Mozilla TLS configuration")
    print()
    print(title)
    if not parsed_command_line.check_against_mozilla_config:
        print("    Disabled; use --mozilla-config={old, intermediate, modern}.\n")
    else:

        print(
            f'    Checking results against Mozilla\'s "{parsed_command_line.check_against_mozilla_config}"'
            f" configuration. See https://ssl-config.mozilla.org/ for more details.\n"
        )
        mozilla_checker = MozillaTlsConfigurationChecker.get_default()
        for server_scan_result in all_server_scan_results:
            try:
                mozilla_checker.check_server(
                    against_config=parsed_command_line.check_against_mozilla_config,
                    server_scan_result=server_scan_result,
                )
                print(f"    {server_scan_result.server_location.display_string}: OK - Compliant.\n")

            except ServerNotCompliantWithMozillaTlsConfiguration as e:
                are_all_servers_compliant = False
                print(f"    {server_scan_result.server_location.display_string}: FAILED - Not compliant.")
                for criteria, error_description in e.issues.items():
                    print(f"        * {criteria}: {error_description}")
                print()

            except ServerScanResultIncomplete:
                are_all_servers_compliant = False
                print(
                    f"    {server_scan_result.server_location.display_string}: ERROR - Scan did not run successfully;"
                    f" review the scan logs above."
                )

    if not are_all_servers_compliant:
        # Return a non-zero error code to signal failure (for example to fail a CI/CD pipeline)
        sys.exit(1)


if __name__ == "__main__":
    main()
